# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Friendli Completion API."""

# pylint: disable=line-too-long, no-name-in-module, too-many-locals

from __future__ import annotations

import json
from typing import List, Literal, Optional, Type, Union, overload

from pydantic import ValidationError

from friendli.errors import InvalidGenerationError
from friendli.schema.api.v1.codegen.completions_pb2 import V1CompletionsRequest
from friendli.schema.api.v1.completions import (
    BeamSearchType,
    Completion,
    CompletionLine,
    TokenSequenceParam,
)
from friendli.sdk.api.base import (
    AsyncGenerationStream,
    AsyncServingAPI,
    GenerationStream,
    ServingAPI,
)
from friendli.utils.compat import model_parse


class Completions(ServingAPI[Type[V1CompletionsRequest]]):
    """Friendli completions API."""

    @property
    def _api_path(self) -> str:
        return "v1/completions"

    @property
    def _method(self) -> str:
        return "POST"

    @property
    def _content_type(self) -> str:
        return (
            "application/protobuf" if self._use_protobuf is None else "application/json"
        )

    @property
    def _request_pb_cls(self) -> Type[V1CompletionsRequest]:
        return V1CompletionsRequest

    @overload
    def create(
        self,
        *,
        stream: Literal[True],
        model: Optional[str] = None,
        prompt: Optional[str] = None,
        tokens: Optional[List[int]] = None,
        timeout_microseconds: Optional[int] = None,
        max_tokens: Optional[int] = None,
        max_total_tokens: Optional[int] = None,
        min_tokens: Optional[int] = None,
        min_total_tokens: Optional[int] = None,
        n: Optional[int] = None,
        num_beams: Optional[int] = None,
        length_penalty: Optional[float] = None,
        early_stopping: Optional[bool] = None,
        no_repeat_ngram: Optional[int] = None,
        encoder_no_repeat_ngram: Optional[int] = None,
        repetition_penalty: Optional[float] = None,
        encoder_repetition_penalty: Optional[float] = None,
        frequency_penalty: Optional[float] = None,
        presence_penalty: Optional[float] = None,
        temperature: Optional[float] = None,
        top_k: Optional[int] = None,
        top_p: Optional[float] = None,
        stop: Optional[List[str]] = None,
        stop_tokens: Optional[List[TokenSequenceParam]] = None,
        seed: Optional[List[int]] = None,
        token_index_to_replace: Optional[List[int]] = None,
        embedding_to_replace: Optional[List[float]] = None,
        beam_search_type: Optional[BeamSearchType] = None,
        beam_compat_pre_normalization: Optional[bool] = None,
        beam_compat_no_post_normalization: Optional[bool] = None,
        bad_words: Optional[List[str]] = None,
        bad_word_tokens: Optional[List[TokenSequenceParam]] = None,
        include_output_logits: Optional[bool] = None,
        include_output_logprobs: Optional[bool] = None,
        forced_output_tokens: Optional[List[int]] = None,
        eos_token: Optional[List[int]] = None,
    ) -> CompletionStream:
        """[skip-doc]."""

    @overload
    def create(
        self,
        *,
        stream: Literal[False],
        model: Optional[str] = None,
        prompt: Optional[str] = None,
        tokens: Optional[List[int]] = None,
        timeout_microseconds: Optional[int] = None,
        max_tokens: Optional[int] = None,
        max_total_tokens: Optional[int] = None,
        min_tokens: Optional[int] = None,
        min_total_tokens: Optional[int] = None,
        n: Optional[int] = None,
        num_beams: Optional[int] = None,
        length_penalty: Optional[float] = None,
        early_stopping: Optional[bool] = None,
        no_repeat_ngram: Optional[int] = None,
        encoder_no_repeat_ngram: Optional[int] = None,
        repetition_penalty: Optional[float] = None,
        encoder_repetition_penalty: Optional[float] = None,
        frequency_penalty: Optional[float] = None,
        presence_penalty: Optional[float] = None,
        temperature: Optional[float] = None,
        top_k: Optional[int] = None,
        top_p: Optional[float] = None,
        stop: Optional[List[str]] = None,
        stop_tokens: Optional[List[TokenSequenceParam]] = None,
        seed: Optional[List[int]] = None,
        token_index_to_replace: Optional[List[int]] = None,
        embedding_to_replace: Optional[List[float]] = None,
        beam_search_type: Optional[BeamSearchType] = None,
        beam_compat_pre_normalization: Optional[bool] = None,
        beam_compat_no_post_normalization: Optional[bool] = None,
        bad_words: Optional[List[str]] = None,
        bad_word_tokens: Optional[List[TokenSequenceParam]] = None,
        include_output_logits: Optional[bool] = None,
        include_output_logprobs: Optional[bool] = None,
        forced_output_tokens: Optional[List[int]] = None,
        eos_token: Optional[List[int]] = None,
    ) -> Completion:
        """[skip-doc]."""

    def create(
        self,
        *,
        stream: bool,
        model: Optional[str] = None,
        prompt: Optional[str] = None,
        tokens: Optional[List[int]] = None,
        timeout_microseconds: Optional[int] = None,
        max_tokens: Optional[int] = None,
        max_total_tokens: Optional[int] = None,
        min_tokens: Optional[int] = None,
        min_total_tokens: Optional[int] = None,
        n: Optional[int] = None,
        num_beams: Optional[int] = None,
        length_penalty: Optional[float] = None,
        early_stopping: Optional[bool] = None,
        no_repeat_ngram: Optional[int] = None,
        encoder_no_repeat_ngram: Optional[int] = None,
        repetition_penalty: Optional[float] = None,
        encoder_repetition_penalty: Optional[float] = None,
        frequency_penalty: Optional[float] = None,
        presence_penalty: Optional[float] = None,
        temperature: Optional[float] = None,
        top_k: Optional[int] = None,
        top_p: Optional[float] = None,
        stop: Optional[List[str]] = None,
        stop_tokens: Optional[List[TokenSequenceParam]] = None,
        seed: Optional[List[int]] = None,
        token_index_to_replace: Optional[List[int]] = None,
        embedding_to_replace: Optional[List[float]] = None,
        beam_search_type: Optional[BeamSearchType] = None,
        beam_compat_pre_normalization: Optional[bool] = None,
        beam_compat_no_post_normalization: Optional[bool] = None,
        bad_words: Optional[List[str]] = None,
        bad_word_tokens: Optional[List[TokenSequenceParam]] = None,
        include_output_logits: Optional[bool] = None,
        include_output_logprobs: Optional[bool] = None,
        forced_output_tokens: Optional[List[int]] = None,
        eos_token: Optional[List[int]] = None,
    ) -> Union[CompletionStream, Completion]:
        """Creates a completion.

        :::note
        #### Compatibility with Hugging Face's beam search

        TL;DR: To mimic the default behavior of Hugging Face's beam search, set `beam_compat_pre_normalization` and `beam_compat_no_post_normalization` to true.

        Our inference API provides following options to match the behavior of Hugging Face's beam search (including both `beam_search` and `beam_sample`).
        In Hugging Face's implementation, normalization of logits occurs _before_ perturbing the logits by applying options like [`no_repeat_ngram_size`](https://huggingface.co/docs/transformers/v4.26.0/en/main_classes/text_generation#transformers.GenerationConfig.no_repeat_ngram_size); this makes Hugging Face's beam scoring procedure use unnormalized logprobs.
        While the Hugging Face team provides an option that performs normalization again after the perturbation (see [`renormalize_logits`](https://huggingface.co/docs/transformers/v4.26.0/en/main_classes/text_generation#transformers.GenerationConfig.renormalize_logits)), they chose not to change their default behavior — you can find related posts [here](https://discuss.huggingface.co/t/announcement-generation-get-probabilities-for-generated-output/30075/14) and [there](https://github.com/huggingface/transformers/pull/19143#issuecomment-1287429478).
        By default, our API applies normalization once after the perturbation.
        If you want to stick to Hugging Face's behavior for compatibility reasons, you should control **beam_compat_pre_normalization** and **beam_compat_no_post_normalization** arguments yourself (both default to false).

        To sum up, you have three choices:

        1. Do not control **beam_compat_pre_normalization** and **beam_compat_no_post_normalization** arguments (i.e., leave both of them as false). Our API will apply normalization once after the perturbation, using the normalized logprobs during beam scoring.
        2. Set both **beam_compat_pre_normalization** and **beam_compat_no_post_normalization** arguments to true. Doing so, the API will mimic Hugging Face's default behavior (i.e., `renormalize_logits=False`), performing normalization once before the perturbation.
        3. Set **beam_compat_pre_normalization** to true (and leave **beam_compat_no_post_normalization** as false). Doing so, the API will mimic Hugging Face's behavior _with_ `renormalize_logits=True`, performing normalization both before and after the perturbation. Note that you cannot set **beam_compat_no_post_normalization** as true when **beam_compat_pre_normalization** is false.
        :::

        Args:
            stream (bool, optional): Whether to stream generation result. When set true, each token will be sent as server-sent events once generated. Not supported when using beam search.
            model (Optional[str]): ID of the model to use. This argument should be set only for serverless endpoints.
            prompt (Optional[str], optional): The prompt (i.e., input text) to generate completion for. Either `prompt` or `tokens` field is required. Defaults to None.
            tokens (Optional[List[int]], optional): The tokenized prompt (i.e., input tokens). Either `prompt` or `tokens` field is required. Defaults to None.
            timeout_microseconds (Optional[int], optional): Request timeout. Gives the HTTP `429 Too Many Requests` response status code. Default behavior is no timeout. Defaults to None.
            max_tokens (Optional[int], optional): The maximum number of tokens to generate. For decoder-only models like GPT, the length of your input tokens plus `max_tokens` should not exceed the model's maximum length (e.g., 2048 for OpenAI GPT-3). For encoder-decoder models like T5 or BlenderBot, `max_tokens` should not exceed the model's maximum output length. This is similar to Hugging Face's [`max_new_tokens`](https://huggingface.co/docs/transformers/v4.26.0/en/main_classes/text_generation#transformers.GenerationConfig.max_new_tokens) argument. Defaults to None.
            max_total_tokens (Optional[int], optional): The maximum number of tokens including both the generated result and the input tokens. Only allowed for decoder-only models. Only one argument between `max_tokens` and `max_total_tokens` is allowed. Default value is the model's maximum length. This is similar to Hugging Face's [`max_length`](https://huggingface.co/docs/transformers/v4.26.0/en/main_classes/text_generation#transformers.GenerationConfig.max_length) argument. Defaults to None.
            min_tokens (Optional[int], optional): The minimum number of tokens to generate. Default value is 0. This is similar to Hugging Face's [`min_new_tokens`](https://huggingface.co/docs/transformers/v4.26.0/en/main_classes/text_generation#transformers.GenerationConfig.min_new_tokens) argument. Defaults to None.
            min_total_tokens (Optional[int], optional): The minimum number of tokens including both the generated result and the input tokens. Only allowed for decoder-only models. Only one argument between `min_tokens` and `min_total_tokens` is allowed. This is similar to Hugging Face's [`min_length`](https://huggingface.co/docs/transformers/v4.26.0/en/main_classes/text_generation#transformers.GenerationConfig.min_length) argument. Defaults to None.
            n (Optional[int], optional): The number of independently generated results for the prompt. Not supported when using beam search. This is similar to Hugging Face's [`num_return_sequences`](https://huggingface.co/docs/transformers/v4.26.0/en/main_classes/text_generation#transformers.GenerationConfig.num_return_sequences(int, ) argument. Defaults to None.
            num_beams (Optional[int], optional): Number of beams for beam search. Numbers between 1 and 31 (both inclusive) are allowed. Default behavior is no beam search. This is similar to Hugging Face's [`num_beams`](https://huggingface.co/docs/transformers/v4.26.0/en/main_classes/text_generation#transformers.GenerationConfig.num_beams) argument. Defaults to None.
            length_penalty (Optional[float], optional): Coefficient for exponential length penalty that is used with beam search. Only allowed for beam search. Defaults to 1.0. This is similar to Hugging Face's [`length_penalty`](https://huggingface.co/docs/transformers/v4.26.0/en/main_classes/text_generation#transformers.GenerationConfig.length_penalty) argument. Defaults to None.
            early_stopping (Optional[bool], optional): Whether to stop the beam search when at least `num_beams` beams are finished with the EOS token. Only allowed for beam search. Defaults to false. This is similar to Hugging Face's [`early_stopping`](https://huggingface.co/docs/transformers/v4.26.0/en/main_classes/text_generation#transformers.GenerationConfig.early_stopping) argument. Defaults to None.
            no_repeat_ngram (Optional[int], optional): If this exceeds 1, every ngram of that size can only occur once among the generated result (plus the input tokens for decoder-only models). 1 means that this mechanism is disabled (i.e., you cannot prevent 1-gram from being generated repeatedly). Defaults to 1. This is similar to Hugging Face's [`no_repeat_ngram_size`](https://huggingface.co/docs/transformers/v4.26.0/en/main_classes/text_generation#transformers.GenerationConfig.no_repeat_ngram_size) argument. Defaults to None.
            encoder_no_repeat_ngram (Optional[int], optional): If this exceeds 1, every ngram of that size occurring in the input token sequence cannot appear in the generated result. 1 means that this mechanism is disabled (i.e., you cannot prevent 1-gram from being generated repeatedly). Only allowed for encoder-decoder models. Defaults to 1. This is similar to Hugging Face's [`encoder_no_repeat_ngram_size`](https://huggingface.co/docs/transformers/v4.26.0/en/main_classes/text_generation#transformers.GenerationConfig.encoder_no_repeat_ngram_size) argument. Defaults to None.
            repetition_penalty (Optional[float], optional): Penalizes tokens that have already appeared in the generated result (plus the input tokens for decoder-only models). Should be greater than or equal to 1.0. 1.0 means no penalty. See [Keskar et al., 2019](https://arxiv.org/abs/1909.05858) for more details. This is similar to Hugging Face's [`repetition_penalty`](https://huggingface.co/docs/transformers/v4.26.0/en/main_classes/text_generation#transformers.GenerationConfig.repetition_penalty) argument. Defaults to None.
            encoder_repetition_penalty (Optional[float], optional): Penalizes tokens that have already appeaared in the input tokens. Should be greater than or equal to 1.0. 1.0 means no penalty. Only allowed for encoder-decoder models. See [Keskar et al., 2019](https://arxiv.org/abs/1909.05858) for more details. This is similar to Hugging Face's [`encoder_repetition_penalty`](https://huggingface.co/docs/transformers/v4.26.0/en/main_classes/text_generation#transformers.GenerationConfig.encoder_repetition_penalty) argument. Defaults to None.
            frequency_penalty (Optional[float], optional): Number between -2.0 and 2.0. Positive values penalizes tokens that have been sampled, taking into account their frequency in the preceding text. This penalization diminishes the model's tendency to reproduce identical lines verbatim. Defaults to None.
            presence_penalty (Optional[float], optional): Number between -2.0 and 2.0. Positive values penalizes tokens that have been sampled at least once in the existing text. Defaults to None.
            temperature (Optional[float], optional): Sampling temperature. Non-zero positive numbers are allowed. Smaller temperature makes the generation result closer to greedy, argmax (i.e., `top_k = 1`) sampling. Defaults to 1.0. This is similar to Hugging Face's [`temperature`](https://huggingface.co/docs/transformers/v4.26.0/en/main_classes/text_generation#transformers.GenerationConfig.temperature) argument. Defaults to None.
            top_k (Optional[int], optional): The number of highest probability tokens to keep for sampling. Numbers between 0 and the vocab size of the model (both inclusive) are allowed. The default value is 0, which means that the API does not apply top-k filtering. This is similar to Hugging Face's [`top_k`](https://huggingface.co/docs/transformers/v4.26.0/en/main_classes/text_generation#transformers.GenerationConfig.top_k) argument. Defaults to None.
            top_p (Optional[float], optional): Tokens comprising the top `top_p` probability mass are kept for sampling. Numbers between 0.0 (exclusive) and 1.0 (inclusive) are allowed. Defaults to 1.0. This is similar to Hugging Face's [`top_p`](https://huggingface.co/docs/transformers/v4.26.0/en/main_classes/text_generation#transformers.GenerationConfig.top_p) argument. Defaults to None.
            stop (Optional[List[str]], optional): When one of the stop phrases appears in the generation result, the API will stop generation. The phrase is included in the generated result. If you are using beam search, all of the active beams should contain the stop phrase to terminate generation. Before checking whether a stop phrase is included in the result, the phrase is converted into tokens. We recommend using `stop_tokens` because it is clearer. For example, after tokenization, phrases "clear" and " clear" can result in different token sequences due to the prepended space character. Defaults to None.
            stop_tokens (Optional[List[TokenSequenceParam]], optional): Same as the above `stop` field, but receives token sequences instead of text phrases. A TokenSequence type is a dict with the key 'tokens' and the value type List[int]. Defaults to None.
            seed (Optional[List[int]], optional): Seed to control random procedure. If nothing is given, the API generate the seed randomly, use it for sampling, and return the seed along with the generated result. When using the `n` argument, you can pass a list of seed values to control all of the independent generations. Defaults to None.
            token_index_to_replace (Optional[List[int]], optional): A list of token indices where to replace the embeddings of input tokens provided via either `tokens` or `prompt`. Defaults to None.
            embedding_to_replace (Optional[List[float]], optional): A list of flattened embedding vectors used for replacing the tokens at the specified indices provided via `token_index_to_replace`. Defaults to None.
            beam_search_type (Optional[BeamSearchType], optional): Which beam search type to use. `DETERMINISTIC` means the standard, deterministic beam search, which is similar to Hugging Face's [`beam_search`](https://huggingface.co/docs/transformers/v4.26.0/en/main_classes/text_generation#transformers.GenerationMixin.beam_search). Argmuents for controlling random sampling such as `top_k` and `top_p` are not allowed for this option. `STOCHASTIC` means stochastic beam search (more details in [Kool et al. (2019)](https://proceedings.mlr.press/v97/kool19a.html)). `NAIVE_SAMPLING` is similar to Hugging Face's [`beam_sample`](https://huggingface.co/docs/transformers/v4.26.0/en/main_classes/text_generation#transformers.GenerationMixin.beam_sample). Defaults to `DETERMINISTIC`. Defaults to None.
            beam_compat_pre_normalization (Optional[bool], optional): Whether to perform normalization both before and after the perturbation. Defaults to None.
            beam_compat_no_post_normalization (Optional[bool], optional): When sets to `True`, do not perform normalization after the perturbation. Defaults to None.
            bad_words (Optional[List[str]], optional): Text phrases that should not be generated. For a bad word phrase that contains N tokens, if the first N-1 tokens appears at the last of the generated result, the logit for the last token of the phrase is set to -inf. We recommend using `bad_word_tokens` because it is clearer (more details in the document for `stop` field). Defaults to empty list. Defaults to None.
            bad_word_tokens (Optional[List[TokenSequenceParam]], optional): Same as the above `bad_words` field, but receives token sequences instead of text phrases. A TokenSequence type is a dict with the key 'tokens' and the value type List[int]. This is similar to Hugging Face's <a href="https://huggingface.co/docs/transformers/v4.26.0/en/main_classes/text_generation#transformers.GenerationConfig.bad_words_ids(List[List[int]]," target="_top">`bad_word_ids`</a> argument. Defaults to None.
            include_output_logits (Optional[bool], optional): Whether to include the output logits to the generation output. Defaults to None.
            include_output_logprobs (Optional[bool], optional): Whether to include the output logprobs to the generation output. Defaults to None.
            forced_output_tokens (Optional[List[int]], optional): A token sequence that is enforced as a generation output. This option can be used when evaluating the model for the datasets with multi-choice problems (e.g., [HellaSwag](https://huggingface.co/datasets/hellaswag), [MMLU](https://huggingface.co/datasets/cais/mmlu)). Use this option with `include_output_logprobs` to get logprobs for the evaluation. Defaults to None.
            eos_token (Optional[List[int]], optional): A list of endpoint sentence tokens. Defaults to None.

        Raises:
            APIError: Raised when the HTTP API request fails.

        Returns:
            Union[Completion, CompletionStream]: If `stream` is `True`, a `CompletionStream` object that iterates the results per token is returned. Otherwise, a `Completion` object is returned.

        Examples:
            Basic usage:

            ```python
            from friendli import Friendli

            client = Friendli()
            completion = client.completions.create(
                prompt="Python is a popular language for",
                stream=False,
                max_tokens=100,
                top_p=0.8,
                temperature=0.5,
                no_repeat_ngram=3,
            )
            print(completion.choices[0].text)
            ```

            Usage of streaming mode:

            ```python
            from friendli import Friendli

            client = Friendli()
            stream = client.completions.create(
                prompt="Python is a popular language for",
                stream=True,  # Enable stream mode.
                max_tokens=100,
                top_p=0.8,
                temperature=0.5,
                no_repeat_ngram=3,
            )

            # Iterate over a generation stream.
            for line in stream:
                print(line.text)

            # Or you can wait for the generation stream to complete.
            completion = stream.wait()
            ```

        """
        request_dict = {
            "stream": stream,
            "prompt": prompt,
            "tokens": tokens,
            "timeout_microseconds": timeout_microseconds,
            "max_tokens": max_tokens,
            "max_total_tokens": max_total_tokens,
            "min_tokens": min_tokens,
            "min_total_tokens": min_total_tokens,
            "n": n,
            "num_beams": num_beams,
            "length_penalty": length_penalty,
            "early_stopping": early_stopping,
            "no_repeat_ngram": no_repeat_ngram,
            "encoder_no_repeat_ngram": encoder_no_repeat_ngram,
            "repetition_penalty": repetition_penalty,
            "encoder_repetition_penalty": encoder_repetition_penalty,
            "frequency_penalty": frequency_penalty,
            "presence_penalty": presence_penalty,
            "temperature": temperature,
            "top_k": top_k,
            "top_p": top_p,
            "stop": stop,
            "stop_tokens": stop_tokens,
            "seed": seed,
            "token_index_to_replace": token_index_to_replace,
            "embedding_to_replace": embedding_to_replace,
            "beam_search_type": beam_search_type,
            "beam_compat_pre_normalization": beam_compat_pre_normalization,
            "beam_compat_no_post_normalization": beam_compat_no_post_normalization,
            "bad_words": bad_words,
            "bad_word_tokens": bad_word_tokens,
            "include_output_logits": include_output_logits,
            "include_output_logprobs": include_output_logprobs,
            "forced_output_tokens": forced_output_tokens,
            "eos_token": eos_token,
        }
        response = self._request(data=request_dict, stream=stream, model=model)

        if stream:
            return CompletionStream(response=response)
        return model_parse(Completion, response.json())


class AsyncCompletions(AsyncServingAPI[Type[V1CompletionsRequest]]):
    """Async completions."""

    @property
    def _api_path(self) -> str:
        return "v1/completions"

    @property
    def _method(self) -> str:
        return "POST"

    @property
    def _content_type(self) -> str:
        return (
            "application/protobuf" if self._use_protobuf is None else "application/json"
        )

    @property
    def _request_pb_cls(self) -> Type[V1CompletionsRequest]:
        return V1CompletionsRequest

    @overload
    async def create(
        self,
        *,
        stream: Literal[True],
        model: Optional[str] = None,
        prompt: Optional[str] = None,
        tokens: Optional[List[int]] = None,
        timeout_microseconds: Optional[int] = None,
        max_tokens: Optional[int] = None,
        max_total_tokens: Optional[int] = None,
        min_tokens: Optional[int] = None,
        min_total_tokens: Optional[int] = None,
        n: Optional[int] = None,
        num_beams: Optional[int] = None,
        length_penalty: Optional[float] = None,
        early_stopping: Optional[bool] = None,
        no_repeat_ngram: Optional[int] = None,
        encoder_no_repeat_ngram: Optional[int] = None,
        repetition_penalty: Optional[float] = None,
        encoder_repetition_penalty: Optional[float] = None,
        frequency_penalty: Optional[float] = None,
        presence_penalty: Optional[float] = None,
        temperature: Optional[float] = None,
        top_k: Optional[int] = None,
        top_p: Optional[float] = None,
        stop: Optional[List[str]] = None,
        stop_tokens: Optional[List[TokenSequenceParam]] = None,
        seed: Optional[List[int]] = None,
        token_index_to_replace: Optional[List[int]] = None,
        embedding_to_replace: Optional[List[float]] = None,
        beam_search_type: Optional[BeamSearchType] = None,
        beam_compat_pre_normalization: Optional[bool] = None,
        beam_compat_no_post_normalization: Optional[bool] = None,
        bad_words: Optional[List[str]] = None,
        bad_word_tokens: Optional[List[TokenSequenceParam]] = None,
        include_output_logits: Optional[bool] = None,
        include_output_logprobs: Optional[bool] = None,
        forced_output_tokens: Optional[List[int]] = None,
        eos_token: Optional[List[int]] = None,
    ) -> AsyncCompletionStream:
        """[skip-doc]."""

    @overload
    async def create(
        self,
        *,
        stream: Literal[False],
        model: Optional[str] = None,
        prompt: Optional[str] = None,
        tokens: Optional[List[int]] = None,
        timeout_microseconds: Optional[int] = None,
        max_tokens: Optional[int] = None,
        max_total_tokens: Optional[int] = None,
        min_tokens: Optional[int] = None,
        min_total_tokens: Optional[int] = None,
        n: Optional[int] = None,
        num_beams: Optional[int] = None,
        length_penalty: Optional[float] = None,
        early_stopping: Optional[bool] = None,
        no_repeat_ngram: Optional[int] = None,
        encoder_no_repeat_ngram: Optional[int] = None,
        repetition_penalty: Optional[float] = None,
        encoder_repetition_penalty: Optional[float] = None,
        frequency_penalty: Optional[float] = None,
        presence_penalty: Optional[float] = None,
        temperature: Optional[float] = None,
        top_k: Optional[int] = None,
        top_p: Optional[float] = None,
        stop: Optional[List[str]] = None,
        stop_tokens: Optional[List[TokenSequenceParam]] = None,
        seed: Optional[List[int]] = None,
        token_index_to_replace: Optional[List[int]] = None,
        embedding_to_replace: Optional[List[float]] = None,
        beam_search_type: Optional[BeamSearchType] = None,
        beam_compat_pre_normalization: Optional[bool] = None,
        beam_compat_no_post_normalization: Optional[bool] = None,
        bad_words: Optional[List[str]] = None,
        bad_word_tokens: Optional[List[TokenSequenceParam]] = None,
        include_output_logits: Optional[bool] = None,
        include_output_logprobs: Optional[bool] = None,
        forced_output_tokens: Optional[List[int]] = None,
        eos_token: Optional[List[int]] = None,
    ) -> Completion:
        """[skip-doc]."""

    async def create(
        self,
        *,
        stream: bool = False,
        model: Optional[str] = None,
        prompt: Optional[str] = None,
        tokens: Optional[List[int]] = None,
        timeout_microseconds: Optional[int] = None,
        max_tokens: Optional[int] = None,
        max_total_tokens: Optional[int] = None,
        min_tokens: Optional[int] = None,
        min_total_tokens: Optional[int] = None,
        n: Optional[int] = None,
        num_beams: Optional[int] = None,
        length_penalty: Optional[float] = None,
        early_stopping: Optional[bool] = None,
        no_repeat_ngram: Optional[int] = None,
        encoder_no_repeat_ngram: Optional[int] = None,
        repetition_penalty: Optional[float] = None,
        encoder_repetition_penalty: Optional[float] = None,
        frequency_penalty: Optional[float] = None,
        presence_penalty: Optional[float] = None,
        temperature: Optional[float] = None,
        top_k: Optional[int] = None,
        top_p: Optional[float] = None,
        stop: Optional[List[str]] = None,
        stop_tokens: Optional[List[TokenSequenceParam]] = None,
        seed: Optional[List[int]] = None,
        token_index_to_replace: Optional[List[int]] = None,
        embedding_to_replace: Optional[List[float]] = None,
        beam_search_type: Optional[BeamSearchType] = None,
        beam_compat_pre_normalization: Optional[bool] = None,
        beam_compat_no_post_normalization: Optional[bool] = None,
        bad_words: Optional[List[str]] = None,
        bad_word_tokens: Optional[List[TokenSequenceParam]] = None,
        include_output_logits: Optional[bool] = None,
        include_output_logprobs: Optional[bool] = None,
        forced_output_tokens: Optional[List[int]] = None,
        eos_token: Optional[List[int]] = None,
    ) -> Union[AsyncCompletionStream, Completion]:
        """Creates a completion asynchronously.

        Args:
            stream (bool, optional): Whether to stream generation result. When set true, each token will be sent as server-sent events once generated. Not supported when using beam search. Defaults to False.
            model (Optional[str]): ID of the model to use. This argument should be set only for serverless endpoints.
            prompt (Optional[str], optional): The prompt (i.e., input text) to generate completion for. Either `prompt` or `tokens` field is required. Defaults to None.
            tokens (Optional[List[int]], optional): The tokenized prompt (i.e., input tokens). Either `prompt` or `tokens` field is required. Defaults to None.
            timeout_microseconds (Optional[int], optional): Request timeout. Gives the HTTP `429 Too Many Requests` response status code. Default behavior is no timeout. Defaults to None.
            max_tokens (Optional[int], optional): The maximum number of tokens to generate. For decoder-only models like GPT, the length of your input tokens plus `max_tokens` should not exceed the model's maximum length (e.g., 2048 for OpenAI GPT-3). For encoder-decoder models like T5 or BlenderBot, `max_tokens` should not exceed the model's maximum output length. This is similar to Hugging Face's [`max_new_tokens`](https://huggingface.co/docs/transformers/v4.26.0/en/main_classes/text_generation#transformers.GenerationConfig.max_new_tokens) argument. Defaults to None.
            max_total_tokens (Optional[int], optional): The maximum number of tokens including both the generated result and the input tokens. Only allowed for decoder-only models. Only one argument between `max_tokens` and `max_total_tokens` is allowed. Default value is the model's maximum length. This is similar to Hugging Face's [`max_length`](https://huggingface.co/docs/transformers/v4.26.0/en/main_classes/text_generation#transformers.GenerationConfig.max_length) argument. Defaults to None.
            min_tokens (Optional[int], optional): The minimum number of tokens to generate. Default value is 0. This is similar to Hugging Face's [`min_new_tokens`](https://huggingface.co/docs/transformers/v4.26.0/en/main_classes/text_generation#transformers.GenerationConfig.min_new_tokens) argument. Defaults to None.
            min_total_tokens (Optional[int], optional): The minimum number of tokens including both the generated result and the input tokens. Only allowed for decoder-only models. Only one argument between `min_tokens` and `min_total_tokens` is allowed. This is similar to Hugging Face's [`min_length`](https://huggingface.co/docs/transformers/v4.26.0/en/main_classes/text_generation#transformers.GenerationConfig.min_length) argument. Defaults to None.
            n (Optional[int], optional): The number of independently generated results for the prompt. Not supported when using beam search. Defaults to 1. This is similar to Hugging Face's [`num_return_sequences`](https://huggingface.co/docs/transformers/v4.26.0/en/main_classes/text_generation#transformers.GenerationConfig.num_return_sequences(int, ) argument. Defaults to None.
            num_beams (Optional[int], optional): Number of beams for beam search. Numbers between 1 and 31 (both inclusive) are allowed. Default behavior is no beam search. This is similar to Hugging Face's [`num_beams`](https://huggingface.co/docs/transformers/v4.26.0/en/main_classes/text_generation#transformers.GenerationConfig.num_beams) argument. Defaults to None.
            length_penalty (Optional[float], optional): Coefficient for exponential length penalty that is used with beam search. Only allowed for beam search. Defaults to 1.0. This is similar to Hugging Face's [`length_penalty`](https://huggingface.co/docs/transformers/v4.26.0/en/main_classes/text_generation#transformers.GenerationConfig.length_penalty) argument. Defaults to None.
            early_stopping (Optional[bool], optional): Whether to stop the beam search when at least `num_beams` beams are finished with the EOS token. Only allowed for beam search. Defaults to false. This is similar to Hugging Face's [`early_stopping`](https://huggingface.co/docs/transformers/v4.26.0/en/main_classes/text_generation#transformers.GenerationConfig.early_stopping) argument. Defaults to None.
            no_repeat_ngram (Optional[int], optional): If this exceeds 1, every ngram of that size can only occur once among the generated result (plus the input tokens for decoder-only models). 1 means that this mechanism is disabled (i.e., you cannot prevent 1-gram from being generated repeatedly). Defaults to 1. This is similar to Hugging Face's [`no_repeat_ngram_size`](https://huggingface.co/docs/transformers/v4.26.0/en/main_classes/text_generation#transformers.GenerationConfig.no_repeat_ngram_size) argument. Defaults to None.
            encoder_no_repeat_ngram (Optional[int], optional): If this exceeds 1, every ngram of that size occurring in the input token sequence cannot appear in the generated result. 1 means that this mechanism is disabled (i.e., you cannot prevent 1-gram from being generated repeatedly). Only allowed for encoder-decoder models. Defaults to 1. This is similar to Hugging Face's [`encoder_no_repeat_ngram_size`](https://huggingface.co/docs/transformers/v4.26.0/en/main_classes/text_generation#transformers.GenerationConfig.encoder_no_repeat_ngram_size) argument. Defaults to None.
            repetition_penalty (Optional[float], optional): Penalizes tokens that have already appeared in the generated result (plus the input tokens for decoder-only models). Should be greater than or equal to 1.0. 1.0 means no penalty. See [Keskar et al., 2019](https://arxiv.org/abs/1909.05858) for more details. This is similar to Hugging Face's [`repetition_penalty`](https://huggingface.co/docs/transformers/v4.26.0/en/main_classes/text_generation#transformers.GenerationConfig.repetition_penalty) argument. Defaults to None.
            encoder_repetition_penalty (Optional[float], optional): Penalizes tokens that have already appeaared in the input tokens. Should be greater than or equal to 1.0. 1.0 means no penalty. Only allowed for encoder-decoder models. See [Keskar et al., 2019](https://arxiv.org/abs/1909.05858) for more details. This is similar to Hugging Face's [`encoder_repetition_penalty`](https://huggingface.co/docs/transformers/v4.26.0/en/main_classes/text_generation#transformers.GenerationConfig.encoder_repetition_penalty) argument. Defaults to None.
            frequency_penalty (Optional[float], optional): Number between -2.0 and 2.0. Positive values penalizes tokens that have been sampled, taking into account their frequency in the preceding text. This penalization diminishes the model's tendency to reproduce identical lines verbatim. Defaults to None.
            presence_penalty (Optional[float], optional): Number between -2.0 and 2.0. Positive values penalizes tokens that have been sampled at least once in the existing text. Defaults to None.
            temperature (Optional[float], optional): Sampling temperature. Non-zero positive numbers are allowed. Smaller temperature makes the generation result closer to greedy, argmax (i.e., `top_k = 1`) sampling. Defaults to 1.0. This is similar to Hugging Face's [`temperature`](https://huggingface.co/docs/transformers/v4.26.0/en/main_classes/text_generation#transformers.GenerationConfig.temperature) argument. Defaults to None.
            top_k (Optional[int], optional): The number of highest probability tokens to keep for sampling. Numbers between 0 and the vocab size of the model (both inclusive) are allowed. The default value is 0, which means that the API does not apply top-k filtering. This is similar to Hugging Face's [`top_k`](https://huggingface.co/docs/transformers/v4.26.0/en/main_classes/text_generation#transformers.GenerationConfig.top_k) argument. Defaults to None.
            top_p (Optional[float], optional): Tokens comprising the top `top_p` probability mass are kept for sampling. Numbers between 0.0 (exclusive) and 1.0 (inclusive) are allowed. Defaults to 1.0. This is similar to Hugging Face's [`top_p`](https://huggingface.co/docs/transformers/v4.26.0/en/main_classes/text_generation#transformers.GenerationConfig.top_p) argument. Defaults to None.
            stop (Optional[List[str]], optional): When one of the stop phrases appears in the generation result, the API will stop generation. The phrase is included in the generated result. If you are using beam search, all of the active beams should contain the stop phrase to terminate generation. Before checking whether a stop phrase is included in the result, the phrase is converted into tokens. We recommend using `stop_tokens` because it is clearer. For example, after tokenization, phrases "clear" and " clear" can result in different token sequences due to the prepended space character. Defaults to empty list. Defaults to None.
            stop_tokens (Optional[List[TokenSequenceParam]], optional): Same as the above `stop` field, but receives token sequences instead of text phrases. A TokenSequence type is a dict with the key 'tokens' and the value type List[int]. Defaults to None.
            seed (Optional[List[int]], optional): Seed to control random procedure. If nothing is given, the API generate the seed randomly, use it for sampling, and return the seed along with the generated result. When using the `n` argument, you can pass a list of seed values to control all of the independent generations. Defaults to None.
            token_index_to_replace (Optional[List[int]], optional): A list of token indices where to replace the embeddings of input tokens provided via either `tokens` or `prompt`. Defaults to None.
            embedding_to_replace (Optional[List[float]], optional): A list of flattened embedding vectors used for replacing the tokens at the specified indices provided via `token_index_to_replace`. Defaults to None.
            beam_search_type (Optional[BeamSearchType], optional): Which beam search type to use. `DETERMINISTIC` means the standard, deterministic beam search, which is similar to Hugging Face's [`beam_search`](https://huggingface.co/docs/transformers/v4.26.0/en/main_classes/text_generation#transformers.GenerationMixin.beam_search). Argmuents for controlling random sampling such as `top_k` and `top_p` are not allowed for this option. `STOCHASTIC` means stochastic beam search (more details in [Kool et al. (2019)](https://proceedings.mlr.press/v97/kool19a.html)). `NAIVE_SAMPLING` is similar to Hugging Face's [`beam_sample`](https://huggingface.co/docs/transformers/v4.26.0/en/main_classes/text_generation#transformers.GenerationMixin.beam_sample). Defaults to `DETERMINISTIC`. Defaults to None.
            beam_compat_pre_normalization (Optional[bool], optional): Whether to perform normalization both before and after the perturbation. Defaults to None.
            beam_compat_no_post_normalization (Optional[bool], optional): When sets to `True`, do not perform normalization after the perturbation. Defaults to None.
            bad_words (Optional[List[str]], optional): Text phrases that should not be generated. For a bad word phrase that contains N tokens, if the first N-1 tokens appears at the last of the generated result, the logit for the last token of the phrase is set to -inf. We recommend using `bad_word_tokens` because it is clearer (more details in the document for `stop` field). Defaults to empty list. Defaults to None.
            bad_word_tokens (Optional[List[TokenSequenceParam]], optional): Same as the above `bad_words` field, but receives token sequences instead of text phrases. A TokenSequence type is a dict with the key 'tokens' and the value type List[int]. This is similar to Hugging Face's <a href="https://huggingface.co/docs/transformers/v4.26.0/en/main_classes/text_generation#transformers.GenerationConfig.bad_words_ids(List[List[int]]," target="_top">`bad_word_ids`</a> argument. Defaults to None.
            include_output_logits (Optional[bool], optional): Whether to include the output logits to the generation output. Defaults to None.
            include_output_logprobs (Optional[bool], optional): Whether to include the output logprobs to the generation output. Defaults to None.
            forced_output_tokens (Optional[List[int]], optional): A token sequence that is enforced as a generation output. This option can be used when evaluating the model for the datasets with multi-choice problems (e.g., [HellaSwag](https://huggingface.co/datasets/hellaswag), [MMLU](https://huggingface.co/datasets/cais/mmlu)). Use this option with `include_output_logprobs` to get logprobs for the evaluation. Defaults to None.
            eos_token (Optional[List[int]], optional): A list of endpoint sentence tokens. Defaults to None.

        Raises:
            APIError: Raised when the HTTP API request fails.
            SessionClosedError: Raised when the client session is not opened with `api_session()`.

        Returns:
            Union[AsyncCompletionStream, Completion]: If `stream` is `True`, a `AsyncCompletionStream` object that iterates the results per token is returned. Otherwise, a `Completion` object is returned.

        Examples:
            Basic usage:

            ```python
            import asyncio
            from friendli import Friendli

            client = Friendli()

            async def main() -> None:
                completion = await client.completions.create(
                    prompt="Python is a popular language for",
                    stream=False,
                    max_tokens=100,
                    top_p=0.8,
                    temperature=0.5,
                    no_repeat_ngram=3,
                )
                print(completion.choices[0].text)

            asyncio.run(main())
            ```

            Usage of streaming mode:

            ```python
            import asyncio
            from friendli import Friendli

            client = Friendli()

            async def main() -> None:
                stream = await client.completions.create(
                    prompt="Python is a popular language for",
                    stream=True,  # Enable stream mode.
                    max_tokens=100,
                    top_p=0.8,
                    temperature=0.5,
                    no_repeat_ngram=3,
                )
                async for line in stream:
                    print(line.choices[0].text, end="")

            asyncio.run(main())
            ```

        """
        request_dict = {
            "stream": stream,
            "prompt": prompt,
            "tokens": tokens,
            "timeout_microseconds": timeout_microseconds,
            "max_tokens": max_tokens,
            "max_total_tokens": max_total_tokens,
            "min_tokens": min_tokens,
            "min_total_tokens": min_total_tokens,
            "n": n,
            "num_beams": num_beams,
            "length_penalty": length_penalty,
            "early_stopping": early_stopping,
            "no_repeat_ngram": no_repeat_ngram,
            "encoder_no_repeat_ngram": encoder_no_repeat_ngram,
            "repetition_penalty": repetition_penalty,
            "encoder_repetition_penalty": encoder_repetition_penalty,
            "frequency_penalty": frequency_penalty,
            "presence_penalty": presence_penalty,
            "temperature": temperature,
            "top_k": top_k,
            "top_p": top_p,
            "stop": stop,
            "stop_tokens": stop_tokens,
            "seed": seed,
            "token_index_to_replace": token_index_to_replace,
            "embedding_to_replace": embedding_to_replace,
            "beam_search_type": beam_search_type,
            "beam_compat_pre_normalization": beam_compat_pre_normalization,
            "beam_compat_no_post_normalization": beam_compat_no_post_normalization,
            "bad_words": bad_words,
            "bad_word_tokens": bad_word_tokens,
            "include_output_logits": include_output_logits,
            "include_output_logprobs": include_output_logprobs,
            "forced_output_tokens": forced_output_tokens,
            "eos_token": eos_token,
        }
        response = await self._request(data=request_dict, stream=stream, model=model)

        if stream:
            return AsyncCompletionStream(response=response)
        return model_parse(Completion, response.json())


class CompletionStream(GenerationStream[CompletionLine]):
    """Completion stream."""

    def __next__(self) -> CompletionLine:  # noqa: D105
        line = next(self._iter)
        while not line:
            line = next(self._iter)

        parsed = json.loads(line.strip("data: "))
        try:
            return model_parse(CompletionLine, parsed)
        except ValidationError as exc:
            try:
                # The last iteration of the stream returns a response with `V1Completion` schema.
                model_parse(Completion, parsed)
                raise StopIteration from exc
            except ValidationError:
                raise InvalidGenerationError(
                    f"Generation result has invalid schema: {str(exc)}"
                ) from exc

    def wait(self) -> Optional[Completion]:
        """Waits for the generation to complete.

        Raises:
            InvalidGenerationError: Raised when the generation result has invalid format.

        Returns:
            Optional[Completion]: The full generation result.

        """
        for line in self._iter:
            if line:
                parsed = json.loads(line.strip("data: "))
                try:
                    # The last iteration of the stream returns a response with `V1Completion` schema.
                    return model_parse(Completion, parsed)
                except ValidationError as exc:
                    try:
                        # Skip the line response.
                        model_parse(CompletionLine, parsed)
                    except ValidationError:
                        raise InvalidGenerationError(
                            f"Generation result has invalid schema: {str(exc)}"
                        ) from exc
        return None


class AsyncCompletionStream(AsyncGenerationStream[CompletionLine]):
    """Asynchronous completion stream."""

    async def __anext__(self) -> CompletionLine:  # noqa: D105
        line = await self._iter.__anext__()
        while not line:
            line = await self._iter.__anext__()

        parsed = json.loads(line.strip("data: "))
        try:
            return model_parse(CompletionLine, parsed)
        except ValidationError as exc:
            try:
                # The last iteration of the stream returns a response with `V1Completion` schema.
                model_parse(Completion, parsed)
                raise StopAsyncIteration from exc
            except ValidationError:
                raise InvalidGenerationError(
                    f"Generation result has invalid schema: {str(exc)}"
                ) from exc

    async def wait(self) -> Optional[Completion]:  # noqa: D105
        """Waits for the generation to complete.

        Raises:
            InvalidGenerationError: Raised when the generation result has invalid format.

        Returns:
            Optional[Completion]: The full generation result.

        """
        async for line in self._iter:
            if line:
                parsed = json.loads(line.strip("data: "))
                try:
                    # The last iteration of the stream returns a response with `V1Completion` schema.
                    return model_parse(Completion, parsed)
                except ValidationError as exc:
                    try:
                        # Skip the line response.
                        model_parse(CompletionLine, parsed)
                    except ValidationError:
                        raise InvalidGenerationError(
                            f"Generation result has invalid schema: {str(exc)}"
                        ) from exc
        return None



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Friendli BloomForCausalLM QuantizerHook."""

# mypy: ignore-errors

from __future__ import annotations

from typing import Any, Dict, Iterator, List, Tuple, Type, cast

import torch
from transformers.models.bloom import (  # type: ignore[import]
    BloomConfig,
    BloomForCausalLM,
)

from friendli.modules.converter.base import OneOfConverter
from friendli.modules.quantizer.schema.config import SmoothQuantConfig
from friendli.modules.quantizer.schema.data import ModuleName, QuantInput, TFQuantInputs
from friendli.modules.quantizer.smoothquant.base import SmoothQuantHook


class SmoothQuantBloomHook(SmoothQuantHook):
    """SmoothQuant Hook for BloomForCausalLM."""

    def __init__(self, quant_config: Dict[str, Any], converter: OneOfConverter):
        """Initialize SmoothQuantBloomHook."""
        super().__init__(quant_config, converter)
        self.num_heads = cast(BloomConfig, converter.config).num_attention_heads
        self.hidden_size = cast(BloomConfig, converter.config).hidden_size
        self.head_size = self.hidden_size // self.num_heads

    def iter_smooth_norm_weights(
        self,
        model: BloomForCausalLM,
    ) -> Iterator[Tuple[List[torch.Tensor], List[torch.Tensor], ModuleName]]:
        """Returns iterator of layernorm's weight and linear layer's weight pr transformer block in BloomForCausalLM."""
        quant_args = cast(SmoothQuantConfig, self.quant_config).smoothquant_args
        for index, decoder_layer in enumerate(model.transformer.h):  # type: ignore[union-attr]
            # [LayerNorm 1] - [ QKV projection ] gets smoothed
            yield (
                [
                    decoder_layer.input_layernorm.weight.data,
                    decoder_layer.input_layernorm.bias.data,
                ],
                [
                    decoder_layer.self_attention.query_key_value.weight.data,
                ],
                f"{self.quantized_layer_prefix}{index}.self_attention.query_key_value",  # the input tensors fed into Q, K, V matrices are identical.
            )
            # [LayerNorm 2] - [ MLP FF 1 ] gets smoothed
            yield (
                [
                    decoder_layer.post_attention_layernorm.weight.data,
                    decoder_layer.post_attention_layernorm.bias.data,
                ],
                [decoder_layer.mlp.dense_h_to_4h.weight.data],  # [OutDim, InDim]
                f"{self.quantized_layer_prefix}{index}.mlp.dense_h_to_4h",
            )
            if quant_args.attn_fc_smoothing:
                yield (
                    [decoder_layer.attn_fc_pre_smoother.scale.data],
                    [decoder_layer.self_attention.dense.weight.data],
                    f"{self.quantized_layer_prefix}{index}.self_attention.dense",
                )
            if quant_args.ff2_smoothing:
                yield (
                    [decoder_layer.ff2_pre_smoother.scale.data],
                    [decoder_layer.mlp.dense_4h_to_h.weight.data],
                    f"{self.quantized_layer_prefix}{index}.mlp.dense_4h_to_h",
                )

    def reshape_qkv_weight(
        self, attn_layer: torch.nn.Module
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """Reshapes the qkv weight in BloomForCausalLM for Quantization."""
        qkv_layer = cast(torch.nn.Linear, attn_layer.query_key_value)
        split_qkv_weight_list = torch.split(qkv_layer.weight, self.head_size, dim=0)
        num_heads = cast(BloomConfig, self.converter.config).num_attention_heads

        [q_weight, k_weight, v_weight] = [
            torch.cat(
                [split_qkv_weight_list[j * 3 + i] for j in range(num_heads)],
                dim=0,
            ).reshape(-1, self.hidden_size)
            for i in range(3)
        ]
        return q_weight, k_weight, v_weight

    def sort_qkv_output_stats(self, max_output_stat: torch.Tensor) -> torch.Tensor:
        """Sort max_output_stas for seperating qkv_layer's output_stats."""
        split_qkv_output_stat = torch.split(max_output_stat, self.head_size)
        qkv_output_stat_list = [
            torch.cat(
                [split_qkv_output_stat[j * 3 + i] for j in range(self.num_heads)],
            )
            for i in range(3)
        ]
        qkv_output_stat = torch.cat(qkv_output_stat_list)
        return qkv_output_stat

    def iter_tf_quant_inputs(self, model: BloomForCausalLM) -> Iterator[TFQuantInputs]:
        """Returns the layers which should be quantized in transformer block of BloomForCausalLM."""
        for index, decoder_layer in enumerate(model.transformer.h):
            self_attn = decoder_layer.self_attention
            q_weight, k_weight, v_weight = self.reshape_qkv_weight(self_attn)
            qkv_weight = torch.cat([q_weight, k_weight, v_weight], dim=0)
            qkv_weight_out_dim = qkv_weight.size(0)
            fc1 = decoder_layer.mlp.dense_h_to_4h
            fc2 = decoder_layer.mlp.dense_4h_to_h

            yield TFQuantInputs(
                layer_index=index,
                block=decoder_layer,
                q=QuantInput(
                    qkv_weight,
                    f"{self.quantized_layer_prefix}{index}.self_attention.query_key_value",
                    0,
                    qkv_weight_out_dim // 3,
                    self.sort_qkv_output_stats,
                ),
                k=QuantInput(
                    qkv_weight,
                    f"{self.quantized_layer_prefix}{index}.self_attention.query_key_value",
                    qkv_weight_out_dim // 3,
                    qkv_weight_out_dim // 3 * 2,
                    self.sort_qkv_output_stats,
                ),
                v=QuantInput(
                    qkv_weight,
                    f"{self.quantized_layer_prefix}{index}.self_attention.query_key_value",
                    qkv_weight_out_dim // 3 * 2,
                    qkv_weight_out_dim,
                    self.sort_qkv_output_stats,
                ),
                attn_fc=QuantInput(
                    self_attn.dense.weight,
                    f"{self.quantized_layer_prefix}{index}.self_attention.dense",
                    None,
                    None,
                ),
                ff1=QuantInput(
                    fc1.weight,
                    f"{self.quantized_layer_prefix}{index}.mlp.dense_h_to_4h",
                    None,
                    None,
                ),
                ff2=QuantInput(
                    fc2.weight,
                    f"{self.quantized_layer_prefix}{index}.mlp.dense_4h_to_h",
                    None,
                    None,
                ),
            )

    def get_linear_layer_types(self) -> Tuple[Type[torch.nn.Module]]:
        """Returns the linear layer types in BloomForCausalLM."""
        return (torch.nn.Linear,)

    def get_attn_fc_layer(self, decoder_layer: torch.nn.Module) -> torch.nn.Linear:
        """Returns the linear layer after attention in the decoder layer."""
        return decoder_layer.self_attention.dense

    def get_ff2_layer(self, decoder_layer: torch.nn.Module) -> torch.nn.Linear:
        """Returns the linear layer after FF1 in the decoder layer."""
        return decoder_layer.mlp.dense_4h_to_h

    def get_tf_blocks(self, model: BloomForCausalLM) -> List[torch.nn.Module]:
        """Returns the decoder layers(transformer blocks) in the model."""
        return list(model.transformer.h)



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Friendli MPTForCausalLM QuantizerHook."""

# mypy: ignore-errors

from __future__ import annotations

from typing import Iterator, List, Tuple, Type, cast

import torch

from friendli.modules.quantizer.schema.config import SmoothQuantConfig
from friendli.modules.quantizer.schema.data import ModuleName, QuantInput, TFQuantInputs
from friendli.modules.quantizer.smoothquant.base import SmoothQuantHook


class SmoothQuantMPTHook(SmoothQuantHook):
    """SmoothQuant Hook for MPTForCausalLM."""

    def iter_smooth_norm_weights(
        self,
        model: torch.nn.Module,
    ) -> Iterator[Tuple[List[torch.Tensor], List[torch.Tensor], ModuleName]]:
        """Returns iterator of layernorm's weight and linear layer's weight per transformer block in MPTForCausalLM."""
        quant_args = cast(SmoothQuantConfig, self.quant_config).smoothquant_args

        for index, decoder_layer in enumerate(
            model.transformer.blocks  # type: ignore[union-attr, arg-type]
        ):
            # [LayerNorm 1] - [ QKV projection ] gets smoothed
            yield (
                [decoder_layer.norm_1.weight.data],
                [decoder_layer.attn.Wqkv.weight.data],
                f"{self.quantized_layer_prefix}{index}.attn.Wqkv",
            )
            # [LayerNorm 2] - [ MLP FF 1 ] gets smoothed
            yield (
                [decoder_layer.norm_2.weight.data],
                [decoder_layer.ffn.up_proj.weight.data],  # [OutDim, InDim]
                f"{self.quantized_layer_prefix}{index}.ffn.up_proj",
            )
            if quant_args.attn_fc_smoothing:
                yield (
                    [decoder_layer.attn_fc_pre_smoother.scale.data],
                    [decoder_layer.attn.out_proj.weight.data],
                    f"{self.quantized_layer_prefix}{index}.attn.out_proj",
                )
            if quant_args.ff2_smoothing:
                yield (
                    [decoder_layer.ff2_pre_smoother.scale.data],
                    [decoder_layer.ffn.down_proj.weight.data],
                    f"{self.quantized_layer_prefix}{index}.ffn.down_proj",
                )

    def iter_tf_quant_inputs(self, model: torch.nn.Module) -> Iterator[TFQuantInputs]:
        """Returns the layers which should be quantized in transformer block of MPTForCausalLM."""
        for index, decoder_layer in enumerate(
            model.transformer.blocks  # type: ignore[union-attr, arg-type]
        ):
            self_attn = decoder_layer.attn
            q_outdim = (
                self.converter.decoder_num_attention_heads
                * self.converter.decoder_head_size
            )
            kv_outdim = (
                self.converter.decoder_num_kv_attention_heads
                * self.converter.decoder_head_size
            )
            qkv_outdim = self_attn.Wqkv.weight.size(0)
            assert qkv_outdim == q_outdim + kv_outdim * 2
            fc1 = decoder_layer.ffn.up_proj
            fc2 = decoder_layer.ffn.down_proj

            yield TFQuantInputs(
                layer_index=index,
                block=decoder_layer,
                q=QuantInput(
                    self_attn.Wqkv.weight,
                    f"{self.quantized_layer_prefix}{index}.attn.Wqkv",
                    0,
                    q_outdim,
                ),
                k=QuantInput(
                    self_attn.Wqkv.weight,
                    f"{self.quantized_layer_prefix}{index}.attn.Wqkv",
                    q_outdim,
                    q_outdim + kv_outdim,
                ),
                v=QuantInput(
                    self_attn.Wqkv.weight,
                    f"{self.quantized_layer_prefix}{index}.attn.Wqkv",
                    q_outdim + kv_outdim,
                    qkv_outdim,
                ),
                attn_fc=QuantInput(
                    self_attn.out_proj.weight,
                    f"{self.quantized_layer_prefix}{index}.attn.out_proj",
                    None,
                    None,
                ),
                ff1=QuantInput(
                    fc1.weight,
                    f"{self.quantized_layer_prefix}{index}.ffn.up_proj",
                    None,
                    None,
                ),
                ff2=QuantInput(
                    fc2.weight,
                    f"{self.quantized_layer_prefix}{index}.ffn.down_proj",
                    None,
                    None,
                ),
            )

    def get_linear_layer_types(self) -> Tuple[Type[torch.nn.Module]]:
        """Returns the linear layer types in MPTForCausalLM."""
        return (torch.nn.Linear,)

    def get_attn_fc_layer(self, decoder_layer: torch.nn.Module) -> torch.nn.Linear:
        """Returns the linear layer after attention in the decoder layer."""
        return decoder_layer.attn.out_proj

    def get_ff2_layer(self, decoder_layer: torch.nn.Module) -> torch.nn.Linear:
        """Returns the linear layer after FF1 in the decoder layer."""
        return decoder_layer.ffn.down_proj

    def get_tf_blocks(self, model: torch.nn.Module) -> List[torch.nn.Module]:
        """Returns the decoder layers(transformer blocks) in the model."""
        return list(model.transformer.blocks)



# Copyright (c) 2024-present, FriendliAI Inc. All rights reserved.

"""Friendli Model GQL Client."""

from __future__ import annotations

from typing import Any, Dict, List

from friendli.client.graphql.base import GqlClient

GetModelListInProjectOp = """
query GetDedicatedModelListInProject($id: ID!) {
  dedicatedProject(id: $id) {
    id
    name
    models {
      totalCount
      edges {
        node {
          name
          id
        }
      }
    }
  }
}
"""


class ModelGqlClient(GqlClient):
    """Model GQL client."""

    def list(self, project_id: str) -> List[Dict[str, Any]]:
        """List models in the project."""
        response = self.run(query=GetModelListInProjectOp, variables={"id": project_id})
        return response["dedicatedProject"]["models"]["edges"]



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Friendli LLaMAa Checkpoint Converter."""


from __future__ import annotations

from typing import Any, Dict, Iterable, List, Set, cast

import torch
from transformers import LlamaConfig, LlamaForCausalLM  # type: ignore[import]

from friendli.errors import CheckpointConversionError, NotSupportedCheckpointError
from friendli.logging import logger
from friendli.modules.converter.base import (
    DECODER_PREFIX,
    DecoderOnlyConverter,
    DecoderOnlyLoraConverter,
)
from friendli.modules.converter.interface import RotaryEmbeddingConversionInterface
from friendli.modules.converter.schema import ConvertInfo
from friendli.modules.converter.utils import convert_to_gpt_j_params


class LlamaForCausalLMLoraConverter(DecoderOnlyLoraConverter):
    """LlamaForCausalLM LoRA Converter Class."""

    def pre_convert(
        self,
        model: torch.nn.Module,
    ) -> torch.nn.Module:
        """Adjust the LoRA Adapter module's params in Llama before converting."""
        converter = cast(LlamaForCausalLMConverter, self.converter)
        for layer in cast(LlamaForCausalLM, model).model.layers:
            query_b = layer.self_attn.q_proj.lora_B.default.weight
            query_b = query_b.reshape(
                converter.decoder_num_attention_heads,
                converter.decoder_head_size,
                -1,
            )
            query_b = convert_to_gpt_j_params(query_b, converter.decoder_head_size)
            query_b = query_b.reshape(
                converter.decoder_num_attention_heads * converter.decoder_head_size,
                -1,
            )
            layer.self_attn.q_proj.lora_B.default.weight.data = query_b
        return model

    @property
    def adapter_target_module_map(self) -> Dict[str, str]:
        """Return the dictionary that maps Hugging Face's module name to Friendli's module name."""
        return {
            "q_proj": "query",
            "k_proj": "key",
            "v_proj": "value",
            "o_proj": "attn_fc",
            "up_proj": "ff1",
            "gate_proj": "ff_gate",
            "down_proj": "ff2",
            "embed_tokens": "wte",
        }

    @property
    def adapter_convert_info_list(
        self,
    ) -> List[ConvertInfo]:
        """The list of conversion informations for LoRA adapter modules in Llama."""
        convert_info_list = []
        target_modules = self.adapter_target_modules

        # Non-transformer modules
        if "wte" in target_modules:
            convert_info_list.extend(
                [
                    ConvertInfo(
                        param_names=["model.embed_tokens.lora_embedding_A.default"],
                        data_type=self.converter.data_type,
                        converted_name="wte/lora/lora_A/weight:0",
                        reshape_fn=self.lora_weight_reshape,
                    ),
                    ConvertInfo(
                        param_names=["model.embed_tokens.lora_embedding_B.default"],
                        data_type=self.converter.data_type,
                        converted_name="wte/lora/lora_B/weight:0",
                        reshape_fn=self.lora_weight_reshape,
                    ),
                ]
            )

        # Transformer modules
        for i in range(self.converter.decoder_layer_num):
            layer_prefix = f"{self.converter.decoder_layer_prefix}{i}."
            converted_prefix = f"{DECODER_PREFIX}/h_._{i}/"
            assert self.adapter_config.target_modules is not None

            if "query" in target_modules:
                convert_info_list.extend(
                    [
                        ConvertInfo(
                            param_names=[
                                f"{layer_prefix}self_attn.q_proj.lora_A.default.weight"
                            ],
                            data_type=self.converter.data_type,
                            converted_name=f"{converted_prefix}attn/c_attn/lora/query_A/weight:0",
                            reshape_fn=self.lora_weight_reshape,
                        ),
                        ConvertInfo(
                            param_names=[
                                f"{layer_prefix}self_attn.q_proj.lora_B.default.weight"
                            ],
                            data_type=self.converter.data_type,
                            converted_name=f"{converted_prefix}attn/c_attn/lora/query_B/weight:0",
                            reshape_fn=self.lora_weight_reshape,
                        ),
                    ]
                )

            if "key" in target_modules:
                convert_info_list.extend(
                    [
                        ConvertInfo(
                            param_names=[
                                f"{layer_prefix}self_attn.k_proj.lora_A.default.weight"
                            ],
                            data_type=self.converter.data_type,
                            converted_name=f"{converted_prefix}attn/c_attn/lora/key_A/weight:0",
                            reshape_fn=self.lora_weight_reshape,
                        ),
                        ConvertInfo(
                            param_names=[
                                f"{layer_prefix}self_attn.k_proj.lora_B.default.weight"
                            ],
                            data_type=self.converter.data_type,
                            converted_name=f"{converted_prefix}attn/c_attn/lora/key_B/weight:0",
                            reshape_fn=self.lora_weight_reshape,
                        ),
                    ]
                )

            if "value" in target_modules:
                convert_info_list.extend(
                    [
                        ConvertInfo(
                            param_names=[
                                f"{layer_prefix}self_attn.v_proj.lora_A.default.weight"
                            ],
                            data_type=self.converter.data_type,
                            converted_name=f"{converted_prefix}attn/c_attn/lora/value_A/weight:0",
                            reshape_fn=self.lora_weight_reshape,
                        ),
                        ConvertInfo(
                            param_names=[
                                f"{layer_prefix}self_attn.v_proj.lora_B.default.weight"
                            ],
                            data_type=self.converter.data_type,
                            converted_name=f"{converted_prefix}attn/c_attn/lora/value_B/weight:0",
                            reshape_fn=self.lora_weight_reshape,
                        ),
                    ]
                )

            if "attn_fc" in target_modules:
                convert_info_list.extend(
                    [
                        ConvertInfo(
                            param_names=[
                                f"{layer_prefix}self_attn.o_proj.lora_A.default.weight"
                            ],
                            data_type=self.converter.data_type,
                            converted_name=f"{converted_prefix}attn/c_proj/lora/lora_A/weight:0",
                            reshape_fn=self.lora_weight_reshape,
                        ),
                        ConvertInfo(
                            param_names=[
                                f"{layer_prefix}self_attn.o_proj.lora_B.default.weight"
                            ],
                            data_type=self.converter.data_type,
                            converted_name=f"{converted_prefix}attn/c_proj/lora/lora_B/weight:0",
                            reshape_fn=self.lora_weight_reshape,
                        ),
                    ]
                )

            if "ff1" in target_modules:
                convert_info_list.extend(
                    [
                        ConvertInfo(
                            param_names=[
                                f"{layer_prefix}mlp.up_proj.lora_A.default.weight"
                            ],
                            data_type=self.converter.data_type,
                            converted_name=f"{converted_prefix}mlp/c_fc/lora/lora_A/weight:0",
                            reshape_fn=self.lora_weight_reshape,
                        ),
                        ConvertInfo(
                            param_names=[
                                f"{layer_prefix}mlp.up_proj.lora_B.default.weight"
                            ],
                            data_type=self.converter.data_type,
                            converted_name=f"{converted_prefix}mlp/c_fc/lora/lora_B/weight:0",
                            reshape_fn=self.lora_weight_reshape,
                        ),
                    ]
                )

            if "ff_gate" in target_modules:
                convert_info_list.extend(
                    [
                        ConvertInfo(
                            param_names=[
                                f"{layer_prefix}mlp.gate_proj.lora_A.default.weight"
                            ],
                            data_type=self.converter.data_type,
                            converted_name=f"{converted_prefix}mlp/c_gate/lora/lora_A/weight:0",
                            reshape_fn=self.lora_weight_reshape,
                        ),
                        ConvertInfo(
                            param_names=[
                                f"{layer_prefix}mlp.gate_proj.lora_B.default.weight"
                            ],
                            data_type=self.converter.data_type,
                            converted_name=f"{converted_prefix}mlp/c_gate/lora/lora_B/weight:0",
                            reshape_fn=self.lora_weight_reshape,
                        ),
                    ]
                )

            if "ff2" in target_modules:
                convert_info_list.extend(
                    [
                        ConvertInfo(
                            param_names=[
                                f"{layer_prefix}mlp.down_proj.lora_A.default.weight"
                            ],
                            data_type=self.converter.data_type,
                            converted_name=f"{converted_prefix}mlp/c_proj/lora/lora_A/weight:0",
                            reshape_fn=self.lora_weight_reshape,
                        ),
                        ConvertInfo(
                            param_names=[
                                f"{layer_prefix}mlp.down_proj.lora_B.default.weight"
                            ],
                            data_type=self.converter.data_type,
                            converted_name=f"{converted_prefix}mlp/c_proj/lora/lora_B/weight:0",
                            reshape_fn=self.lora_weight_reshape,
                        ),
                    ]
                )

        return convert_info_list


class LlamaForCausalLMConverter(
    DecoderOnlyConverter, RotaryEmbeddingConversionInterface
):
    """LlamaForCausalLM Architectures Converter Class."""

    def check_config(self) -> None:
        """Check if LLaMA architectures' config can be converted to Friendli format."""
        super().check_config()
        try:
            if cast(LlamaConfig, self.config).hidden_act not in ["silu"]:
                raise NotSupportedCheckpointError(
                    invalid_option=f"'hidden_act={cast(LlamaConfig, self.config).hidden_act}'",
                    valid_options=["silu"],
                )
            if cast(LlamaConfig, self.config).tie_word_embeddings:
                raise NotSupportedCheckpointError(
                    invalid_option="'tie_word_embeddings=True'",
                    valid_options=[False],
                )
            if cast(LlamaConfig, self.config).rms_norm_eps not in (1e-5, 1e-6):
                raise NotSupportedCheckpointError(
                    invalid_option=f"'rms_norm_eps={cast(LlamaConfig, self.config).rms_norm_eps}'",
                    valid_options=[1e-5, 1e-6],
                )
        except AttributeError as exc:
            raise CheckpointConversionError(str(exc)) from exc

    def qkv_weight_reshape(self, params: List[torch.Tensor]) -> torch.Tensor:
        """qkv_weight_reshape for LLaMA's attention layer."""
        assert len(params) == 3
        q_weight = params[0]
        k_weight = params[1]
        v_weight = params[2]

        q_weight = q_weight.reshape(
            self.decoder_num_attention_heads,
            self.decoder_head_size,
            self.decoder_hidden_size,
        )
        k_weight = k_weight.reshape(
            self.decoder_num_kv_attention_heads,
            self.decoder_head_size,
            self.decoder_hidden_size,
        )
        q_weight = convert_to_gpt_j_params(q_weight, self.rotary_dim)
        k_weight = convert_to_gpt_j_params(k_weight, self.rotary_dim)
        q_weight = q_weight.reshape(
            self.decoder_num_attention_heads * self.decoder_head_size,
            self.decoder_hidden_size,
        )
        k_weight = k_weight.reshape(
            self.decoder_num_kv_attention_heads * self.decoder_head_size,
            self.decoder_hidden_size,
        )

        qkv_weight = torch.cat([q_weight, k_weight, v_weight], dim=0)
        qkv_weight = qkv_weight.transpose(0, -1)
        return qkv_weight

    def get_attributes(self) -> Dict[str, Any]:
        """Get checkpoint attributes."""
        config = cast(LlamaConfig, self.config)

        logger.info(
            "The generated attributes set 'max_length' to %d, but you can change the "
            "'max_length' according to your needs. The Llama model does not rely on "
            "absolute position embeddings, allowing you to choose any suitable value.",
            config.max_position_embeddings,
        )

        eos_token_id = self.get_eos_token_id()
        attr = {
            "model_type": self.model_type,
            "dtype": self.data_type.value,
            "head_size": self.decoder_head_size,
            "rotary_dim": self.rotary_dim,
            "num_heads": self.decoder_num_attention_heads,
            "num_kv_heads": self.decoder_num_kv_attention_heads,
            "num_layers": self.decoder_layer_num,
            "ff_intermediate_size": self.decoder_ff_intermediate_size,
            "max_length": config.max_position_embeddings,
            "vocab_size": config.vocab_size,
            "eos_token": eos_token_id if eos_token_id is not None else "FILL ME",
            "rope_theta": self.rotary_emb_base,
        }
        return attr

    @property
    def model_type(self) -> str:
        """Model type."""
        return "llama"

    @property
    def decoder_convert_info_list(
        self,
    ) -> List[ConvertInfo]:
        """The list of conversion informations for transformer blocks in LLaMA."""
        convert_info_list = []
        for i in range(self.decoder_layer_num):
            layer_prefix = f"{self.decoder_layer_prefix}{i}."
            converted_prefix = f"{DECODER_PREFIX}/h_._{i}/"
            convert_info_list.extend(
                [
                    ConvertInfo(
                        param_names=[f"{layer_prefix}input_layernorm.weight"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}ln_1/gamma:0",
                        reshape_fn=self.ln_weight_reshape,
                    ),
                    ConvertInfo(
                        param_names=[
                            f"{layer_prefix}self_attn.q_proj.weight",
                            f"{layer_prefix}self_attn.k_proj.weight",
                            f"{layer_prefix}self_attn.v_proj.weight",
                        ],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}attn/c_attn/weight:0",
                        reshape_fn=self.qkv_weight_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}self_attn.o_proj.weight"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}attn/c_proj/weight:0",
                        reshape_fn=self.linear_weight_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}post_attention_layernorm.weight"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}ln_2/gamma:0",
                        reshape_fn=self.ln_weight_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}mlp.gate_proj.weight"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}mlp/c_gate/weight:0",
                        reshape_fn=self.linear_weight_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}mlp.up_proj.weight"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}mlp/c_fc/weight:0",
                        reshape_fn=self.linear_weight_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}mlp.down_proj.weight"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}mlp/c_proj/weight:0",
                        reshape_fn=self.linear_weight_reshape,
                    ),
                ]
            )
        return convert_info_list

    @property
    def non_transformer_convert_info_list(
        self,
    ) -> List[ConvertInfo]:
        """The list of conversion informations for non-transformer blocks in LLaMA."""
        return [
            ConvertInfo(
                param_names=["model.embed_tokens.weight"],
                data_type=self.data_type,
                converted_name="wte/weight:0",
                reshape_fn=self.token_embed_weight_reshape,
            ),
            ConvertInfo(
                param_names=["model.norm.weight"],
                data_type=self.data_type,
                converted_name=f"{DECODER_PREFIX}/ln_f/gamma:0",
                reshape_fn=self.ln_weight_reshape,
            ),
            ConvertInfo(
                param_names=["lm_head.weight"],
                data_type=self.data_type,
                converted_name=f"head_fc/weight:0",
                reshape_fn=self.head_weight_reshape,
            ),
        ]

    @property
    def decoder_layer_prefix(self) -> str:
        """The layer name prefix used before LLaMA's transformer block number."""
        return "model.layers."

    @property
    def decoder_layer_num(self) -> int:
        """The number of decoder layers in LLaMA."""
        return cast(LlamaConfig, self.config).num_hidden_layers

    @property
    def decoder_hidden_size(self) -> int:
        """The hidden size in LLaMA."""
        return cast(LlamaConfig, self.config).hidden_size

    @property
    def decoder_num_attention_heads(self) -> int:
        """The number of attention heads in LLaMA."""
        return cast(LlamaConfig, self.config).num_attention_heads

    @property
    def decoder_num_kv_attention_heads(self) -> int:
        """The number of key-value attention heads in LLaMA."""
        config = cast(LlamaConfig, self.config)
        if config.num_key_value_heads is None:
            return self.decoder_num_attention_heads
        return config.num_key_value_heads

    @property
    def decoder_head_size(self) -> int:
        """The head size of LLaMA."""
        return self.decoder_hidden_size // self.decoder_num_attention_heads

    @property
    def decoder_ff_intermediate_size(self) -> int:
        """The intermediate size of the linear layer in LLaMA MLP."""
        return self.config.intermediate_size

    @property
    def rotary_dim(self) -> int:
        """The rotary embedding dimension of LLaMA."""
        return self.decoder_head_size

    @property
    def rotary_emb_base(self) -> float:
        """The rotary embedding base of LLaMA."""
        return cast(LlamaConfig, self.config).rope_theta



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Friendli CLI Interactive Prompt Utilities."""

from __future__ import annotations

import os
from subprocess import CalledProcessError, check_call
from typing import Optional

import typer


def open_editor(path: str, editor: Optional[str] = None):
    """Open the terminal editor."""
    default_editor = editor or get_default_editor()
    try:
        check_call([default_editor, path])
    except CalledProcessError:
        typer.secho("", fg=typer.colors.RED)


def get_default_editor() -> str:
    """Get default terminal editor."""
    return os.environ.get("FRIENDLI_CLI_EDITOR", "vim")



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Friendli MPT Checkpoint Converter."""

from __future__ import annotations

from typing import Any, Dict, List, cast

from transformers import (  # type: ignore[import]
    GenerationConfig,
    MptConfig,
    PretrainedConfig,
)

from friendli.enums import ModelDataType  # type: ignore[import]
from friendli.errors import CheckpointConversionError, NotSupportedCheckpointError
from friendli.logging import logger
from friendli.modules.converter.base import (
    DECODER_PREFIX,
    DecoderOnlyConverter,
    DecoderOnlyLoraConverter,
)
from friendli.modules.converter.schema import ConvertInfo


def safe_attn_config_get(attn_config: Dict[str, Any], key: str) -> Any:
    """Safe getter from MptAttentionConfig.

    This function is a temporary function because MptAttentionConfig
    is not supported `attn_type="grouped_query_attention"` yet.
    """
    if key not in attn_config:
        raise CheckpointConversionError(
            f"{key} does not exist in MptAttentionConfig {attn_config}"
        )

    return attn_config[key]


class MptForCausalLMLoraConverter(DecoderOnlyLoraConverter):
    """MptForCausalLM LoRA Converter Class."""

    @property
    def adapter_target_module_map(self) -> Dict[str, str]:
        """Return the dictionary that maps Hugging Face's module name to Friendli's module name."""
        return {
            "Wqkv": "merged-qkv",
            "out_proj": "attn_fc",
            "up_proj": "ff1",
            "down_proj": "ff2",
            "wte": "wte",
        }

    @property
    def adapter_convert_info_list(
        self,
    ) -> List[ConvertInfo]:
        """The list of conversion informations for LoRA adapter modules in Mpt."""
        convert_info_list = []
        target_modules = self.adapter_target_modules

        # Non-transformer modules
        if "wte" in target_modules:
            convert_info_list.extend(
                [
                    ConvertInfo(
                        param_names=["transformer.wte.lora_embedding_A.default"],
                        data_type=self.converter.data_type,
                        converted_name="wte/lora/lora_A/weight:0",
                        reshape_fn=self.lora_weight_reshape,
                    ),
                    ConvertInfo(
                        param_names=["transformer.wte.lora_embedding_B.default"],
                        data_type=self.converter.data_type,
                        converted_name="wte/lora/lora_B/weight:0",
                        reshape_fn=self.lora_weight_reshape,
                    ),
                ]
            )

        # Transformer modules
        for i in range(self.converter.decoder_layer_num):
            layer_prefix = f"{self.converter.decoder_layer_prefix}{i}."
            converted_prefix = f"{DECODER_PREFIX}/h_._{i}/"

            if "merged-qkv" in target_modules:
                convert_info_list.extend(
                    [
                        ConvertInfo(
                            param_names=[
                                f"{layer_prefix}attn.Wqkv.lora_A.default.weight"
                            ],
                            data_type=self.converter.data_type,
                            converted_name=f"{converted_prefix}attn/c_attn/lora/lora_A/weight:0",
                            reshape_fn=self.lora_weight_reshape,
                        ),
                        ConvertInfo(
                            param_names=[
                                f"{layer_prefix}attn.Wqkv.lora_B.default.weight"
                            ],
                            data_type=self.converter.data_type,
                            converted_name=f"{converted_prefix}attn/c_attn/lora/lora_B/weight:0",
                            reshape_fn=self.lora_weight_reshape,
                        ),
                    ]
                )

            if "attn_fc" in target_modules:
                convert_info_list.extend(
                    [
                        ConvertInfo(
                            param_names=[
                                f"{layer_prefix}attn.out_proj.lora_A.default.weight"
                            ],
                            data_type=self.converter.data_type,
                            converted_name=f"{converted_prefix}attn/c_proj/lora/lora_A/weight:0",
                            reshape_fn=self.lora_weight_reshape,
                        ),
                        ConvertInfo(
                            param_names=[
                                f"{layer_prefix}attn.out_proj.lora_B.default.weight"
                            ],
                            data_type=self.converter.data_type,
                            converted_name=f"{converted_prefix}attn/c_proj/lora/lora_B/weight:0",
                            reshape_fn=self.lora_weight_reshape,
                        ),
                    ]
                )

            if "ff1" in target_modules:
                convert_info_list.extend(
                    [
                        ConvertInfo(
                            param_names=[
                                f"{layer_prefix}mlp.up_proj.lora_A.default.weight"
                            ],
                            data_type=self.converter.data_type,
                            converted_name=f"{converted_prefix}mlp/c_fc/lora/lora_A/weight:0",
                            reshape_fn=self.lora_weight_reshape,
                        ),
                        ConvertInfo(
                            param_names=[
                                f"{layer_prefix}mlp.up_proj.lora_B.default.weight"
                            ],
                            data_type=self.converter.data_type,
                            converted_name=f"{converted_prefix}mlp/c_fc/lora/lora_B/weight:0",
                            reshape_fn=self.lora_weight_reshape,
                        ),
                    ]
                )

            if "ff2" in target_modules:
                convert_info_list.extend(
                    [
                        ConvertInfo(
                            param_names=[
                                f"{layer_prefix}mlp.down_proj.lora_A.default.weight"
                            ],
                            data_type=self.converter.data_type,
                            converted_name=f"{converted_prefix}mlp/c_proj/lora/lora_A/weight:0",
                            reshape_fn=self.lora_weight_reshape,
                        ),
                        ConvertInfo(
                            param_names=[
                                f"{layer_prefix}mlp.down_proj.lora_B.default.weight"
                            ],
                            data_type=self.converter.data_type,
                            converted_name=f"{converted_prefix}mlp/c_proj/lora/lora_B/weight:0",
                            reshape_fn=self.lora_weight_reshape,
                        ),
                    ]
                )

        return convert_info_list


class MPTForCausalLMConverter(DecoderOnlyConverter):
    """MPTForCausalLM Architectures Converter Class."""

    def __init__(
        self,
        config: PretrainedConfig,
        generation_config: GenerationConfig | None,
        data_type: ModelDataType,
    ) -> None:
        """Initialize MPTForCausalLMConverter."""
        super().__init__(config, generation_config, data_type)
        attn_config = cast(MptConfig, config).attn_config
        if isinstance(attn_config, PretrainedConfig):
            attn_config = attn_config.to_dict()  # type: ignore
        self.attn_config = attn_config

    def check_config(self) -> None:
        """Check if MPT architectures' config can be converted to Friendli format."""
        super().check_config()

        if not safe_attn_config_get(self.attn_config, "alibi"):
            raise NotSupportedCheckpointError(
                invalid_option=f"'alibi={safe_attn_config_get(self.attn_config, 'alibi')}'",
                valid_options=[True],
            )

        if safe_attn_config_get(self.attn_config, "alibi_bias_max") != 8:
            raise NotSupportedCheckpointError(
                invalid_option=f"'alibi={safe_attn_config_get(self.attn_config, 'alibi_bias_max')}'",
                valid_options=[8],
            )

        if safe_attn_config_get(self.attn_config, "attn_type") != "multihead_attention":
            if (
                safe_attn_config_get(self.attn_config, "attn_type")
                == "grouped_query_attention"
            ):
                raise CheckpointConversionError(
                    msg="MptAttentionConfig does not support `attn_type=`grouped_query_attention`` yet (as of transformers==4.35.2).",
                )
            raise NotSupportedCheckpointError(
                invalid_option=f"'attn_type={safe_attn_config_get(self.attn_config, 'attn_type')}'",
                valid_options=["multihead_attention"],
            )

        if safe_attn_config_get(self.attn_config, "prefix_lm"):
            raise NotSupportedCheckpointError(
                invalid_option=f"'prefix_lm={safe_attn_config_get(self.attn_config, 'prefix_lm')}'",
                valid_options=[False],
            )

        if safe_attn_config_get(self.attn_config, "qk_ln"):
            raise NotSupportedCheckpointError(
                invalid_option=f"'qk_ln={safe_attn_config_get(self.attn_config, 'qk_ln')}'",
                valid_options=[False],
            )

        if safe_attn_config_get(self.attn_config, "softmax_scale") is not None:
            raise NotSupportedCheckpointError(
                invalid_option=f"'softmax_scale={safe_attn_config_get(self.attn_config, 'softmax_scale')}'",
                valid_options=[None],
            )

        if cast(MptConfig, self.config).expansion_ratio != 4:
            raise NotSupportedCheckpointError(
                invalid_option=(
                    f"'expansion_ratio={cast(MptConfig, self.config).expansion_ratio}'"
                ),
                valid_options=[4],
            )

        if not cast(MptConfig, self.config).no_bias:
            raise NotSupportedCheckpointError(
                invalid_option=f"'no_bias={cast(MptConfig, self.config).no_bias}'",
                valid_options=[True],
            )

        if cast(MptConfig, self.config).logit_scale is not None:
            raise NotSupportedCheckpointError(
                invalid_option=(
                    f"'logit_scale={cast(MptConfig, self.config).logit_scale}'"
                ),
                valid_options=[None],
            )

    @property
    def decoder_convert_info_list(
        self,
    ) -> List[ConvertInfo]:
        """The list of conversion informations for transformer blocks in MPT."""
        convert_info_list = []
        for i in range(self.decoder_layer_num):
            layer_prefix = f"{self.decoder_layer_prefix}{i}."
            converted_prefix = f"{DECODER_PREFIX}/h_._{i}/"
            convert_info_list.extend(
                [
                    ConvertInfo(
                        param_names=[f"{layer_prefix}norm_1.weight"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}ln_1/gamma:0",
                        reshape_fn=self.ln_weight_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}norm_2.weight"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}ln_2/gamma:0",
                        reshape_fn=self.ln_weight_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}attn.Wqkv.weight"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}attn/c_attn/weight:0",
                        reshape_fn=self.qkv_weight_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}attn.out_proj.weight"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}attn/c_proj/weight:0",
                        reshape_fn=self.linear_weight_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}ffn.up_proj.weight"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}mlp/c_fc/weight:0",
                        reshape_fn=self.linear_weight_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}ffn.down_proj.weight"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}mlp/c_proj/weight:0",
                        reshape_fn=self.linear_weight_reshape,
                    ),
                ]
            )

        return convert_info_list

    @property
    def non_transformer_convert_info_list(
        self,
    ) -> List[ConvertInfo]:
        """The list of conversion informations for non-transformer blocks in MPT."""
        return [
            ConvertInfo(
                param_names=["transformer.wte.weight"],
                data_type=self.data_type,
                converted_name="wte/weight:0",
                reshape_fn=self.token_embed_weight_reshape,
            ),
            ConvertInfo(
                param_names=["transformer.norm_f.weight"],
                data_type=self.data_type,
                converted_name=f"{DECODER_PREFIX}/ln_f/gamma:0",
                reshape_fn=self.ln_weight_reshape,
            ),
        ]

    def get_attributes(self) -> Dict[str, Any]:
        """Get checkpoint attributes."""
        logger.info(
            "The generated attributes set 'max_length' to %d, but you can change the "
            "'max_length' according to your needs. The MPT model does not rely on "
            "absolute position embeddings, allowing you to choose any suitable value.",
            cast(MptConfig, self.config).max_seq_len,
        )

        attr = {
            "model_type": self.model_type,
            "dtype": self.data_type.value,
            "head_size": self.decoder_head_size,
            "num_heads": self.decoder_num_attention_heads,
            "num_kv_heads": self.decoder_num_kv_attention_heads,
            "num_layers": self.decoder_layer_num,
            "max_length": cast(MptConfig, self.config).max_seq_len,
            "vocab_size": cast(MptConfig, self.config).vocab_size,
            "clip_qkv": safe_attn_config_get(self.attn_config, "clip_qkv") or 0.0,
            "eos_token": self.get_eos_token_id() or "FILL ME",
        }
        return attr

    @property
    def model_type(self) -> str:
        """Model type."""
        return "mpt"

    @property
    def decoder_layer_prefix(self) -> str:
        """The layer name prefix used before the MPT's transformer block number."""
        return "transformer.blocks."

    @property
    def decoder_layer_num(self) -> int:
        """The number of decoder layers in MPT."""
        return cast(MptConfig, self.config).n_layers

    @property
    def decoder_hidden_size(self) -> int:
        """The hidden size in MPT."""
        return cast(MptConfig, self.config).d_model

    @property
    def decoder_num_attention_heads(self) -> int:
        """The number of attention heads in MPT."""
        return cast(MptConfig, self.config).n_heads

    @property
    def decoder_num_kv_attention_heads(self) -> int:
        """The number of key-value attention heads in MPT."""
        if "kv_n_heads" in self.attn_config:
            return self.attn_config["kv_n_heads"]
        return self.decoder_num_attention_heads

    @property
    def decoder_head_size(self) -> int:
        """The head size of MPT."""
        return self.decoder_hidden_size // self.decoder_num_attention_heads

    @property
    def decoder_ff_intermediate_size(self) -> int:
        """The intermediate size of the linear layer in MPT MLP."""
        return self.decoder_hidden_size * 4



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Friendli GPTJForCausalLM QuantizerHook."""

# mypy: ignore-errors

from __future__ import annotations

import copy
from typing import Iterator, List, Tuple, Type, cast

import torch
from transformers.models.gptj import GPTJForCausalLM  # type: ignore[import]

from friendli.modules.converter.base import DECODER_PREFIX
from friendli.modules.converter.schema import ConvertInfo
from friendli.modules.quantizer.schema.config import SmoothQuantConfig
from friendli.modules.quantizer.schema.data import ModuleName, QuantInput, TFQuantInputs
from friendli.modules.quantizer.smoothquant.base import SmoothQuantHook


class SmoothQuantGPTJHook(SmoothQuantHook):
    """SmoothQuant Hook for GPTJForCausalLM."""

    def pre_smooth(self, model: torch.nn.Module) -> torch.nn.Module:
        """Pre-procedure for SmoothQuant in GPTJForCausalLM that should be called before smooth() is called."""
        super().pre_smooth(model)
        for decoder_layer in cast(GPTJForCausalLM, model).transformer.h:
            decoder_layer.add_module("ln_2", copy.deepcopy(decoder_layer.ln_1))
        return model

    def iter_smooth_norm_weights(
        self,
        model: GPTJForCausalLM,
    ) -> Iterator[Tuple[List[torch.Tensor], List[torch.Tensor], ModuleName]]:
        """Returns iterator of layernorm's weight and linear layer's weight per transformer block in GPTJForCausalLM."""
        quant_args = cast(SmoothQuantConfig, self.quant_config).smoothquant_args
        for index, decoder_layer in enumerate(model.transformer.h):  # type: ignore[union-attr]
            # [LayerNorm 1] - [ QKV projection] gets smoothed
            yield (
                [
                    decoder_layer.ln_1.weight.data,
                    decoder_layer.ln_1.bias.data,
                ],
                [
                    decoder_layer.attn.q_proj.weight.data,  # [OutDim, InDim]
                    decoder_layer.attn.k_proj.weight.data,  # [OutDim, InDim]
                    decoder_layer.attn.v_proj.weight.data,  # [OutDim, InDim]
                ],
                f"{self.quantized_layer_prefix}{index}.attn.q_proj",  # the input tensors fed into Q, K, V matrices are identical.
            )
            # [LayerNorm 1] - [ MLP FF1 ] gets smoothed
            yield (
                [
                    decoder_layer.ln_2.weight.data,
                    decoder_layer.ln_2.bias.data,
                ],
                [
                    decoder_layer.mlp.fc_in.weight.data,
                ],
                f"{self.quantized_layer_prefix}{index}.mlp.fc_in",
            )
            if quant_args.attn_fc_smoothing:
                yield (
                    [decoder_layer.attn_fc_pre_smoother.scale.data],
                    [decoder_layer.attn.out_proj.weight.data],
                    f"{self.quantized_layer_prefix}{index}.attn.out_proj",
                )
            if quant_args.ff2_smoothing:
                yield (
                    [decoder_layer.ff2_pre_smoother.scale.data],
                    [decoder_layer.mlp.fc_out.weight.data],
                    f"{self.quantized_layer_prefix}{index}.mlp.fc_out",
                )

    def iter_tf_quant_inputs(self, model: GPTJForCausalLM) -> Iterator[TFQuantInputs]:
        """Returns the layers which should be quantized in transformer block of GPTJForCausalLM."""
        for index, decoder_layer in enumerate(model.transformer.h):
            attn = decoder_layer.attn
            fc1 = decoder_layer.mlp.fc_in
            fc2 = decoder_layer.mlp.fc_out
            yield TFQuantInputs(
                layer_index=index,
                block=decoder_layer,
                q=QuantInput(
                    attn.q_proj.weight,
                    f"{self.quantized_layer_prefix}{index}.attn.q_proj",
                    None,
                    None,
                ),
                k=QuantInput(
                    attn.k_proj.weight,
                    f"{self.quantized_layer_prefix}{index}.attn.k_proj",
                    None,
                    None,
                ),
                v=QuantInput(
                    attn.v_proj.weight,
                    f"{self.quantized_layer_prefix}{index}.attn.v_proj",
                    None,
                    None,
                ),
                attn_fc=QuantInput(
                    attn.out_proj.weight,
                    f"{self.quantized_layer_prefix}{index}.attn.out_proj",
                    None,
                    None,
                ),
                ff1=QuantInput(
                    fc1.weight,
                    f"{self.quantized_layer_prefix}{index}.mlp.fc_in",
                    None,
                    None,
                ),
                ff2=QuantInput(
                    fc2.weight,
                    f"{self.quantized_layer_prefix}{index}.mlp.fc_out",
                    None,
                    None,
                ),
            )

    @property
    def modified_layers_convert_info_list(
        self,
    ) -> List[ConvertInfo]:
        """Returns the modified layers' convert dict in GPTJForCausalLM."""
        convert_info_list = super().modified_layers_convert_info_list

        for i in range(self.converter.decoder_layer_num):
            layer_prefix = f"{self.quantized_layer_prefix}{i}."
            converted_prefix = f"{DECODER_PREFIX}/h_._{i}/"
            convert_info_list.extend(
                [
                    ConvertInfo(
                        param_names=[f"{layer_prefix}ln_2.weight"],
                        data_type=self.converter.data_type,
                        converted_name=f"{converted_prefix}ln_2/gamma:0",
                        reshape_fn=self.converter.ln_weight_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}ln_2.bias"],
                        data_type=self.converter.data_type,
                        converted_name=f"{converted_prefix}ln_2/beta:0",
                        reshape_fn=self.converter.ln_bias_reshape,
                    ),
                ]
            )

        return convert_info_list

    def get_linear_layer_types(self) -> Tuple[Type[torch.nn.Module]]:
        """Returns the linear layer types in GPTJForCausalLM."""
        return (torch.nn.Linear,)

    def get_attn_fc_layer(self, decoder_layer: torch.nn.Module) -> torch.nn.Linear:
        """Returns the linear layer after attention in the decoder layer."""
        return decoder_layer.attn.out_proj

    def get_ff2_layer(self, decoder_layer: torch.nn.Module) -> torch.nn.Linear:
        """Returns the linear layer after FF1 in the decoder layer."""
        return decoder_layer.mlp.fc_out

    def get_tf_blocks(self, model: GPTJForCausalLM) -> List[torch.nn.Module]:
        """Returns the decoder layers(transformer blocks) in the model."""
        return list(model.transformer.h)



# Copyright (c) 2024-present, FriendliAI Inc. All rights reserved.

# pylint: disable=too-many-arguments

"""Generates image from text via CLI."""

from __future__ import annotations

from typing import Optional

import typer

from friendli.enums import ResponseFormat
from friendli.sdk.client import Friendli
from friendli.utils.compat import model_dump
from friendli.utils.decorator import check_api

app = typer.Typer(
    no_args_is_help=True,
    context_settings={"help_option_names": ["-h", "--help"]},
    add_completion=True,
)


@app.command()
@check_api
def create(
    model: str = typer.Option(
        ...,
        "--model",
        "-m",
        help=(
            "The model to use for chat completions. "
            "See https://docs.friendli.ai/guides/serverless_endpoints/pricing for more "
            "about available models and pricing."
        ),
    ),
    prompt: str = typer.Option(
        ...,
        "--prompt",
        "-p",
        help="A text description of the desired image(s).",
    ),
    negative_prompt: Optional[str] = typer.Option(
        None,
        "--negative-prompt",
        "-N",
        help="A text specifying what you don't want in your image(s).",
    ),
    num_outputs: Optional[int] = typer.Option(
        None,
        "--num-outputs",
        "-n",
        min=1,
        max=16,
        help="The number of images to generate. Only 1 output will be generated when not provided.",
    ),
    num_inference_steps: Optional[int] = typer.Option(
        None,
        "--num-inference-steps",
        "-I",
        min=1,
        max=500,
        help=(
            "The number of inference steps for denoising process. 50 steps will be "
            "taken when not provided."
        ),
    ),
    guidance_scale: Optional[float] = typer.Option(
        None,
        "--guidance-scale",
        "-G",
        help=(
            "Guidance scale to control how much generation process adheres to the text "
            "prompt. When not provided, it is set to 7.5."
        ),
    ),
    seed: Optional[int] = typer.Option(
        None,
        "--seed",
        "-s",
        help="Seed to control random procedure.",
    ),
    response_format: ResponseFormat = typer.Option(
        ResponseFormat.URL,
        "--response-format",
        "-R",
        help="The format in which the generated images are returned.",
    ),
    token: Optional[str] = typer.Option(
        None, "--token", "-t", help="Personal access token for auth."
    ),
    team_id: Optional[str] = typer.Option(
        None, "--team-id", help="ID of team to run as."
    ),
):
    """Create image from text."""
    client = Friendli(token=token, team_id=team_id)

    image = client.images.text_to_image.create(
        model=model,
        prompt=prompt,
        negative_prompt=negative_prompt,
        num_outputs=num_outputs,
        num_inference_steps=num_inference_steps,
        guidance_scale=guidance_scale,
        seed=seed,
        response_format=response_format,
    )
    typer.echo(model_dump(image))



# Copyright (c) 2024-present, FriendliAI Inc. All rights reserved.

"""Login Client."""

from __future__ import annotations

from typing import Tuple

from friendli.client.http.base import HttpClient


class LoginClient(HttpClient):
    """Login client."""

    @property
    def url_path(self) -> str:
        """Get an URL path."""
        return self.url_provider.get_web_backend_uri("/api/auth/login")

    def login(self, email: str, pwd: str) -> Tuple[str, str]:
        """Send request to sign in with email and password."""
        payload = {
            "email": email,
            "password": pwd,
        }
        headers = {
            "Accept": "application/json",
            "rid": "anti-csrf",
            "st-auth-mode": "header",
        }
        resp = self.bare_post(json=payload, headers=headers)
        access_token = resp.headers["st-access-token"]
        refresh_token = resp.headers["st-refresh-token"]
        return access_token, refresh_token



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Friendli Client: CLI and SDK for the fastest generative AI serving."""

from __future__ import annotations

import os

from friendli.di.injector import set_default_modules
from friendli.di.modules import default_modules
from friendli.sdk.client import AsyncFriendli, Friendli

token = os.environ.get("FRIENDLI_TOKEN")
team_id = os.environ.get("FRIENDLI_TEAM")
project_id = os.environ.get("FRIENDLI_PROJECT")

set_default_modules(default_modules)

__all__ = [
    "token",
    "team_id",
    "project_id",
    "AsyncFriendli",
    "Friendli",
]



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Friendli CLI Output Formatter."""

from __future__ import annotations

import json
import os
import re
from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Dict, List, Optional, Union, cast

from rich import box
from rich.console import Console, RenderableType
from rich.filesize import decimal
from rich.json import JSON
from rich.panel import Panel
from rich.table import Table
from rich.text import Text
from rich.tree import Tree


def get_value(data: Dict[str, Any], keys: str) -> str:
    """Get value of `key` from `data`.

    Unlike dict.get method, it is available to access the nested value in `data`.

    Args:
        data (Dict[str, Any]): Data to read.
        key (str): Dot(.)-separated nested keys in data to access the value.

    Returns:
        str: The retrieved data.

    """
    value: Any = data
    key_list = keys.split(".")
    for key in key_list:
        # e.g. `data.results[2].a`
        getitem_match = re.match(r"(.+)\[(-?\d+)\]$", key)
        if getitem_match:
            value = value.get(getitem_match.group(1))[int(getitem_match.group(2))]
        elif value is not None:
            value = value.get(key)
    if isinstance(value, Enum):
        value = value.value
    if isinstance(value, list):
        value = ", ".join(value)
    if value is None:
        return "-"
    return str(value)


@dataclass
class Formatter:
    """Formatter helps data formatting and visualization."""

    name: str

    def __post_init__(self) -> None:
        """Post-init formatter."""
        self._console = Console()


@dataclass
class ListFormatter(Formatter):
    """Base interface for list data formatter."""

    fields: List[str]
    headers: List[str]
    extra_fields: List[str] = field(default_factory=list)
    extra_headers: List[str] = field(default_factory=list)
    substitute_exact_match_only: bool = True

    def __post_init__(self) -> None:
        """Post-init formatter."""
        super().__post_init__()
        assert len(self.fields) == len(self.headers)
        assert len(self.extra_fields) == len(self.extra_headers)

        self._styling_map: Dict[str, Any] = {}
        self._substitution_rule: Dict[str, Any] = {}

    def render(self, data: List[Dict[str, Any]], show_detail: bool = False) -> None:
        """Print the rendered output."""
        raise NotImplementedError  # pragma: no cover

    def get_renderable(
        self, data: List[Dict[str, Any]], show_detail: bool = False
    ) -> RenderableType:
        """Get rendered visualizer."""
        raise NotImplementedError  # pragma: no cover

    def apply_styling(self, header: str, **kwargs) -> None:
        """Apply the styling."""
        self._styling_map[header] = kwargs

    def add_substitution_rule(self, before: str, after: Any) -> None:
        """Add a substitution rule."""
        self._substitution_rule[before] = after

    def _substitute(self, val: str) -> str:
        if self.substitute_exact_match_only:
            # Substitute only when `val` exactly matches to a rule.
            if val in self._substitution_rule:
                return self._substitution_rule[val]
        else:
            # Apply substitution for all matched substrings.
            for before, after in self._substitution_rule.items():
                if before in val:
                    return val.replace(before, after)
        return val


@dataclass
class TableFormatter(ListFormatter):
    """Table formatter for visualizing tabulated data."""

    caption: Optional[str] = None
    table: Optional[Table] = None

    def _init(self, show_detail: bool) -> None:
        self.table = Table(title=self.name, caption=self.caption, box=box.SIMPLE)
        self._make_header(show_detail)

    def render(self, data: List[Dict[str, Any]], show_detail: bool = False) -> None:
        """Print the rendered output."""
        self._build_table(data, show_detail)
        self._console.print(self.table)

    def get_renderable(
        self, data: List[Dict[str, Any]], show_detail: bool = False
    ) -> Table:
        """Get rendered visualizer."""
        self._build_table(data, show_detail)
        return cast(Table, self.table)

    def _build_table(self, data: List[Dict[str, Any]], show_detail: bool) -> None:
        self._init(show_detail)

        for d in data:
            info = [self._substitute(get_value(d, f)) for f in self.fields]
            if show_detail:
                info.extend(
                    [self._substitute(get_value(d, f)) for f in self.extra_fields]
                )
            cast(Table, self.table).add_row(*info)

    def _make_header(self, show_detail: bool) -> None:
        for header in self.headers:
            cast(Table, self.table).add_column(
                header, **self._styling_map.get(header, {}), overflow="fold"
            )

        if show_detail:
            for extra_header in self.extra_headers:
                cast(Table, self.table).add_column(extra_header)


@dataclass
class PanelFormatter(ListFormatter):
    """Panel formatter for visualizing information detail in a panel."""

    subtitle: Optional[str] = None
    panel: Optional[Panel] = None

    def render(
        self,
        data: Union[List[Dict[str, Any]], Dict[str, Any]],
        show_detail: bool = False,
    ) -> None:
        """Print the rendered output."""
        if not isinstance(data, list):
            data = [data]
        self._build_panel(data, show_detail)
        self._console.print(self.panel)

    def get_renderable(
        self,
        data: Union[List[Dict[str, Any]], Dict[str, Any]],
        show_detail: bool = False,
    ) -> Panel:
        """Get rendered visualizer."""
        if not isinstance(data, list):
            data = [data]
        self._build_panel(data, show_detail)
        return cast(Panel, self.panel)

    def _build_panel(self, data: List[Dict[str, Any]], show_detail: bool):
        headers = self.headers + self.extra_headers if show_detail else self.headers
        table = Table(box=None, show_header=False)
        table.add_column("k", style="dim bold")
        table.add_column("v")

        for d in data:
            info = [self._substitute(get_value(d, f)) for f in self.fields]
            if show_detail:
                info.extend(
                    [self._substitute(get_value(d, f)) for f in self.extra_fields]
                )
            for k, v in zip(headers, info):
                table.add_row(k, v)
        self.panel = Panel(table, title=self.name, subtitle=self.subtitle)


@dataclass
class Edge:
    """Edge of file traversal tree."""

    name: str
    size: int | None


def find_and_insert(parent: Tree, edges: List[Edge]) -> None:
    """Recursively construct a file tree."""
    if not edges:
        return

    is_dir = len(edges) > 1
    folder_style_prefix = "[bold magenta]:open_file_folder: "

    match = [
        tree
        for tree in parent.children
        if edges[0].name
        == str(tree.label).replace(folder_style_prefix, "").split(" ", maxsplit=1)[0]
    ]

    if match:
        tree = match[0]
    else:
        if is_dir:
            tree = parent.add(f"{folder_style_prefix}{edges[0].name}")
        else:
            text_filename = Text(edges[0].name, "green")
            file_size = edges[0].size or 0
            text_filename.append(f" ({decimal(file_size)})", "blue")
            tree = parent.add(text_filename)

    find_and_insert(tree, edges[1:])


class TreeFormatter(Formatter):
    """Tree formatter for visualizing file tree."""

    def __init__(self, name: str, root: str = ""):
        """Initialize tree formatter."""
        super().__init__(name=name)
        self._root = "/" + root.lstrip("/")
        self._panel: Optional[Panel] = None

    def render(self, data: List[Dict[str, Any]]) -> None:
        """Print the rendered output."""
        self._build_renderable(data)
        self._console.print(self._panel)

    def get_renderable(self, data: List[Dict[str, Any]]) -> Panel:
        """Get rendered visualizer."""
        self._build_renderable(data)
        return cast(Panel, self._panel)

    def _build_tree(self, data: List[Dict[str, Any]]) -> Tree:
        root = Tree("/")
        paths = [
            f"{d['path']}"
            if os.path.isabs(d["path"])
            else f"{os.path.join(self._root, d['path'])}"
            for d in data
        ]
        sizes = [d["size"] for d in data]
        for path, size in zip(paths, sizes):
            edges = []
            parts = path.split("/")[1:]
            for i, part in enumerate(parts):
                file_size = size if i == len(parts) - 1 else None
                edges.append(Edge(part, file_size))
            find_and_insert(root, edges)
        return root

    def _build_renderable(self, data: List[Dict[str, Any]]) -> None:
        root = self._build_tree(data)
        self._panel = Panel(root, title=self.name)


@dataclass
class JSONFormatter(Formatter):
    """JSON formatter for visualizing JSON data."""

    panel: Optional[Panel] = None

    def render(self, data: Dict[str, Any]) -> None:
        """Print the rendered output."""
        self._build_json(data)
        self._console.print(self.panel)

    def get_renderable(self, data: Dict[str, Any]) -> Panel:
        """Get rendered visualizer."""
        self._build_json(data)
        return cast(Panel, self.panel)

    def _build_json(self, data: Dict[str, Any]) -> None:
        self.panel = Panel(JSON(json.dumps(data)), title=self.name)



# Copyright (c) 2023-present, FriendliAI Inc. All rights reserved.

"""Friendli Phi Checkpoint Converter."""


from __future__ import annotations

import math
from typing import Any, Dict, List, Optional, cast

import torch
from transformers import PretrainedConfig  # type: ignore[import]

from friendli.errors import CheckpointConversionError, NotSupportedCheckpointError
from friendli.logging import logger
from friendli.modules.converter.base import (
    DECODER_PREFIX,
    SUPPORTED_GELU_FAMILY,
    DecoderOnlyConverter,
)
from friendli.modules.converter.interface import RotaryEmbeddingConversionInterface
from friendli.modules.converter.schema import ConvertInfo
from friendli.modules.converter.utils import convert_to_gpt_j_params


class PhiMsftConfig(PretrainedConfig):
    """Phi msft configuration. Different from the HuggingFace PhiConfig."""

    model_type = "phi"
    attribute_map = {
        "max_position_embeddings": "n_positions",
        "hidden_size": "n_embd",
        "num_attention_heads": "n_head",
        "num_hidden_layers": "n_layer",
    }

    def __init__(
        self,
        vocab_size: int = 50304,
        n_positions: int = 2048,
        n_embd: int = 1024,
        n_layer: int = 20,
        n_inner: Optional[int] = None,
        n_head: int = 16,
        n_head_kv: Optional[int] = None,
        rotary_dim: Optional[int] = 32,
        activation_function: Optional[str] = "gelu_new",
        flash_attn: bool = False,
        flash_rotary: bool = False,
        fused_dense: bool = False,
        attn_pdrop: float = 0.0,
        embd_pdrop: float = 0.0,
        resid_pdrop: float = 0.0,
        layer_norm_epsilon: float = 1e-5,
        initializer_range: float = 0.02,
        tie_word_embeddings: bool = False,
        pad_vocab_size_multiple: int = 64,
        **kwargs,
    ) -> None:
        """Initalize the configuration for a phi-msft model."""
        self.vocab_size = int(
            math.ceil(vocab_size / pad_vocab_size_multiple) * pad_vocab_size_multiple
        )
        self.n_positions = n_positions
        self.n_embd = n_embd
        self.n_layer = n_layer
        self.n_inner = n_inner
        self.n_head = n_head
        self.n_head_kv = n_head_kv
        self.rotary_dim = min(rotary_dim, n_embd // n_head)  # type: ignore[type-var]
        self.activation_function = activation_function
        self.flash_attn = flash_attn
        self.flash_rotary = flash_rotary
        self.fused_dense = fused_dense
        self.attn_pdrop = attn_pdrop
        self.embd_pdrop = embd_pdrop
        self.resid_pdrop = resid_pdrop
        self.layer_norm_epsilon = layer_norm_epsilon
        self.initializer_range = initializer_range

        super().__init__(tie_word_embeddings=tie_word_embeddings, **kwargs)


class PhiForCausalLMConverter(DecoderOnlyConverter):
    """PhiForCausalLM Architectures Converter Class."""

    def check_config(self) -> None:
        """Check if Phi architectures' config can be converted to Friendli format."""
        super().check_config()
        try:
            if (
                cast(PhiMsftConfig, self.config).activation_function
                not in SUPPORTED_GELU_FAMILY
            ):
                raise NotSupportedCheckpointError(
                    invalid_option="'activation_function="
                    f"{cast(PhiMsftConfig, self.config).activation_function}'",
                    valid_options=SUPPORTED_GELU_FAMILY,
                )
            if cast(PhiMsftConfig, self.config).tie_word_embeddings:
                raise NotSupportedCheckpointError(
                    invalid_option="'tie_word_embeddings=True'",
                    valid_options=[False],
                )
        except AttributeError as exc:
            raise CheckpointConversionError(str(exc)) from exc

    def qkv_weight_reshape(
        self,
        params: List[torch.Tensor],
    ) -> torch.Tensor:
        """qkv_weight_reshape for Phi's attention layer."""
        assert len(params) == 1
        qkv_weight = params[0]

        q_size = self.decoder_num_attention_heads * self.decoder_head_size
        kv_size = self.decoder_num_kv_attention_heads * self.decoder_head_size
        q_weight, k_weight, v_weight = torch.split(
            qkv_weight, [q_size, kv_size, kv_size], dim=0
        )

        q_weight = q_weight.reshape(
            self.decoder_num_attention_heads,
            self.decoder_head_size,
            self.decoder_hidden_size,
        )
        k_weight = k_weight.reshape(
            self.decoder_num_kv_attention_heads,
            self.decoder_head_size,
            self.decoder_hidden_size,
        )

        q_weight = convert_to_gpt_j_params(q_weight, self.rotary_dim)
        k_weight = convert_to_gpt_j_params(k_weight, self.rotary_dim)

        q_weight = q_weight.reshape(
            self.decoder_num_attention_heads * self.decoder_head_size,
            self.decoder_hidden_size,
        )
        k_weight = k_weight.reshape(
            self.decoder_num_kv_attention_heads * self.decoder_head_size,
            self.decoder_hidden_size,
        )

        qkv_weight = torch.cat([q_weight, k_weight, v_weight], dim=0)
        qkv_weight = qkv_weight.transpose(0, -1)
        return qkv_weight

    def qkv_bias_reshape(self, params: List[torch.Tensor]) -> torch.Tensor:
        """qkv_bias_reshape for Phi's attention layer."""
        assert len(params) == 1
        qkv_bias = params[0]

        q_size = self.decoder_num_attention_heads * self.decoder_head_size
        kv_size = self.decoder_num_kv_attention_heads * self.decoder_head_size

        q_bias, k_bias, v_bias = torch.split(
            qkv_bias, [q_size, kv_size, kv_size], dim=0
        )

        q_bias = q_bias.reshape(
            self.decoder_num_attention_heads, self.decoder_head_size
        )
        k_bias = k_bias.reshape(
            self.decoder_num_kv_attention_heads, self.decoder_head_size
        )

        q_bias = convert_to_gpt_j_params(q_bias, self.rotary_dim).flatten()
        k_bias = convert_to_gpt_j_params(k_bias, self.rotary_dim).flatten()

        qkv_bias = torch.cat((q_bias, k_bias, v_bias), dim=0)
        return qkv_bias

    def get_attributes(self) -> Dict[str, Any]:
        """Get checkpoint attributes."""
        config = cast(PhiMsftConfig, self.config)

        logger.info(
            "The generated attributes set 'max_length' to %d, but you can change the "
            "'max_length' according to your needs. The Phi model does not rely on "
            "absolute position embeddings, allowing you to choose any suitable value.",
            config.n_positions,
        )

        eos_token_id = self.get_eos_token_id()
        attr = {
            "model_type": self.model_type,
            "dtype": self.data_type.value,
            "head_size": self.decoder_head_size,
            "rotary_dim": self.rotary_dim,
            "num_heads": self.decoder_num_attention_heads,
            "num_kv_heads": self.decoder_num_kv_attention_heads,
            "num_layers": self.decoder_layer_num,
            "ff_intermediate_size": self.decoder_ff_intermediate_size,
            "max_length": config.n_positions,
            "vocab_size": config.vocab_size,
            "eos_token": eos_token_id if eos_token_id is not None else "FILL ME",
            "rope_theta": self.rotary_emb_base,
        }
        return attr

    @property
    def model_type(self) -> str:
        """Model type."""
        return "phi"

    @property
    def decoder_convert_info_list(
        self,
    ) -> List[ConvertInfo]:
        """The list of conversion informations for transformer blocks in Phi."""
        convert_info_list = []
        for i in range(self.decoder_layer_num):
            layer_prefix = f"{self.decoder_layer_prefix}{i}."
            converted_prefix = f"{DECODER_PREFIX}/h_._{i}/"
            convert_info_list.extend(
                [
                    ConvertInfo(
                        param_names=[f"{layer_prefix}ln.weight"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}ln_1/gamma:0",
                        reshape_fn=self.ln_weight_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}ln.bias"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}ln_1/beta:0",
                        reshape_fn=self.ln_bias_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}mlp.fc1.weight"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}mlp/c_fc/weight:0",
                        reshape_fn=self.linear_weight_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}mlp.fc1.bias"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}mlp/c_fc/bias:0",
                        reshape_fn=self.linear_bias_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}mlp.fc2.weight"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}mlp/c_proj/weight:0",
                        reshape_fn=self.linear_weight_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}mlp.fc2.bias"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}mlp/c_proj/bias:0",
                        reshape_fn=self.linear_bias_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}mixer.Wqkv.weight"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}attn/c_attn/weight:0",
                        reshape_fn=self.qkv_weight_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}mixer.Wqkv.bias"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}attn/c_attn/bias:0",
                        reshape_fn=self.qkv_bias_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}mixer.out_proj.weight"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}attn/c_proj/weight:0",
                        reshape_fn=self.linear_weight_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}mixer.out_proj.bias"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}attn/c_proj/bias:0",
                        reshape_fn=self.linear_bias_reshape,
                    ),
                ]
            )
        return convert_info_list

    @property
    def non_transformer_convert_info_list(
        self,
    ) -> List[ConvertInfo]:
        """The list of conversion informations for non-transformer blocks in Phi."""
        return [
            ConvertInfo(
                param_names=["transformer.embd.wte.weight"],
                data_type=self.data_type,
                converted_name="wte/weight:0",
                reshape_fn=self.token_embed_weight_reshape,
            ),
            ConvertInfo(
                param_names=["lm_head.ln.weight"],
                data_type=self.data_type,
                converted_name=f"{DECODER_PREFIX}/ln_f/gamma:0",
                reshape_fn=self.ln_weight_reshape,
            ),
            ConvertInfo(
                param_names=["lm_head.ln.bias"],
                data_type=self.data_type,
                converted_name=f"{DECODER_PREFIX}/ln_f/beta:0",
                reshape_fn=self.ln_bias_reshape,
            ),
            ConvertInfo(
                param_names=["lm_head.linear.weight"],
                data_type=self.data_type,
                converted_name="head_fc/weight:0",
                reshape_fn=self.head_weight_reshape,
            ),
            ConvertInfo(
                param_names=["lm_head.linear.bias"],
                data_type=self.data_type,
                converted_name="head_fc/bias:0",
                reshape_fn=self.head_weight_reshape,
            ),
        ]

    @property
    def decoder_layer_prefix(self) -> str:
        """The layer name prefix used before Phi's transformer module number."""
        return "transformer.h."

    @property
    def decoder_layer_num(self) -> int:
        """The number of decoder layers in Phi."""
        return cast(PhiMsftConfig, self.config).n_layer

    @property
    def decoder_hidden_size(self) -> int:
        """The hidden size in Phi."""
        return cast(PhiMsftConfig, self.config).n_embd

    @property
    def decoder_num_attention_heads(self) -> int:
        """The number of attention heads in Phi."""
        return cast(PhiMsftConfig, self.config).n_head

    @property
    def decoder_num_kv_attention_heads(self) -> int:
        """The number of key-value attention heads in Phi."""
        config = cast(PhiMsftConfig, self.config)
        if config.n_head_kv is not None:
            return config.n_head_kv
        return self.decoder_num_attention_heads

    @property
    def decoder_head_size(self) -> int:
        """The head size of Phi."""
        return self.decoder_hidden_size // self.decoder_num_attention_heads

    @property
    def decoder_ff_intermediate_size(self) -> int:
        """The intermediate size of the linear layer in codegen MLP."""
        config = cast(PhiMsftConfig, self.config)
        if config.n_inner is None:
            return self.decoder_hidden_size * 4
        return config.n_inner

    @property
    def rotary_dim(self) -> int:
        """The rotary dim in Phi."""
        return cast(PhiMsftConfig, self.config).rotary_dim  # type: ignore[return-value]

    @property
    def rotary_emb_base(self) -> float:
        """The rotary emb base in Phi."""
        return 10000.0



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Friendli Serving API Interface."""

# pylint: disable=no-name-in-module

from __future__ import annotations

import json
import os
from abc import ABC, abstractmethod
from typing import Any, Dict, Generic, Optional, Type, TypeVar, Union

import httpx
from google.protobuf import json_format
from pydantic import BaseModel
from typing_extensions import Self

from friendli.auth import get_auth_header
from friendli.errors import APIError
from friendli.schema.api.v1.codegen.chat_completions_pb2 import V1ChatCompletionsRequest
from friendli.schema.api.v1.codegen.completions_pb2 import V1CompletionsRequest
from friendli.schema.api.v1.codegen.text_to_image_pb2 import V1TextToImageRequest
from friendli.utils.request import DEFAULT_REQ_TIMEOUT

_GenerationLine = TypeVar("_GenerationLine", bound=BaseModel)


class GenerationStream(ABC, Generic[_GenerationLine]):
    """Generation stream."""

    def __init__(self, response: httpx.Response) -> None:
        """Initializes generation stream."""
        self._iter = response.iter_lines()

    def __iter__(self) -> Self:  # noqa: D105
        return self

    @abstractmethod
    def __next__(self) -> _GenerationLine:
        """Iterates the stream."""


class AsyncGenerationStream(ABC, Generic[_GenerationLine]):
    """Asynchronous generation stream."""

    def __init__(self, response: httpx.Response) -> None:
        """Initializes generation stream."""
        self._iter = response.aiter_lines()

    def __aiter__(self) -> Self:  # noqa: D105
        return self

    @abstractmethod
    async def __anext__(self) -> _GenerationLine:
        """Iterates the stream."""


_HttpxClient = TypeVar("_HttpxClient", bound=Union[httpx.Client, httpx.AsyncClient])
_ProtoMsgType = TypeVar(
    "_ProtoMsgType",
    bound=Union[
        Type[V1CompletionsRequest],
        Type[V1ChatCompletionsRequest],
        Type[V1TextToImageRequest],
    ],
)


class BaseAPI(ABC, Generic[_HttpxClient, _ProtoMsgType]):
    """Base API interface."""

    _client: _HttpxClient

    def __init__(
        self,
        base_url: str,
        endpoint_id: Optional[str] = None,
        use_protobuf: bool = False,
    ) -> None:
        """Initializes BaseAPI."""
        self._endpoint_id = endpoint_id
        self._host = httpx.URL(base_url)
        self._use_protobuf = use_protobuf

    @property
    @abstractmethod
    def _api_path(self) -> str:
        """API URL path."""

    @property
    @abstractmethod
    def _method(self) -> str:
        """API call method."""

    @property
    @abstractmethod
    def _content_type(self) -> str:
        """Request content type."""

    @property
    @abstractmethod
    def _request_pb_cls(self) -> _ProtoMsgType:
        """Protobuf message class to serialize the data of request body."""

    def _build_request(
        self, data: dict[str, Any], model: Optional[str] = None
    ) -> httpx.Request:
        """Build request."""
        return self._client.build_request(
            method=self._method,
            url=self._build_url(),
            content=self._build_content(data, model),
            files=self._build_files(data),
            headers=self._get_headers(),
            timeout=DEFAULT_REQ_TIMEOUT,
        )

    def _build_url(self) -> httpx.URL:
        path = ""
        if self._endpoint_id is not None:
            path = "dedicated"
        path = os.path.join(path, self._api_path)
        return self._host.join(path)

    def _get_headers(self) -> Dict[str, Any]:
        return {
            "Content-Type": self._content_type,
            **get_auth_header(),
        }

    def _build_files(self, data: dict[str, Any]) -> dict[str, Any] | None:
        if self._content_type.startswith("multipart/form-data"):
            files = {}
            for key, val in data.items():
                if val is not None:
                    files[key] = (None, val)
            return files
        return None

    def _build_content(
        self, data: dict[str, Any], model: Optional[str] = None
    ) -> bytes | None:
        if self._endpoint_id is not None:
            data["model"] = self._endpoint_id
        else:
            data["model"] = model

        if self._content_type.startswith("multipart/form-data"):
            return None

        if self._use_protobuf:
            pb_cls = self._request_pb_cls
            request_pb = pb_cls()
            json_format.ParseDict(data, request_pb)
            return request_pb.SerializeToString()

        return json.dumps(data).encode()


class ServingAPI(BaseAPI[httpx.Client, _ProtoMsgType]):
    """Serving API interface."""

    def __init__(
        self,
        base_url: str,
        endpoint_id: Optional[str] = None,
        use_protobuf: bool = False,
        client: Optional[httpx.Client] = None,
    ) -> None:
        """Initializes ServingAPI."""
        super().__init__(
            base_url=base_url, endpoint_id=endpoint_id, use_protobuf=use_protobuf
        )
        self._client = client or httpx.Client()

    def _request(
        self, *, data: dict[str, Any], stream: bool, model: Optional[str] = None
    ) -> httpx.Response:
        # TODO: Add retry / handle timeout and etc.
        if (
            self._host == "https://inference.friendli.ai"
            and self._endpoint_id is None
            and model is None
        ):
            raise ValueError("`model` is required for serverless endpoints.")
        if self._endpoint_id is not None and model is not None:
            raise ValueError("`model` is not allowed for dedicated endpoints.")

        request = self._build_request(data=data, model=model)
        response = self._client.send(request=request, stream=stream)
        self._check_http_error(response)

        return response

    def _check_http_error(self, response: httpx.Response) -> None:
        try:
            response.raise_for_status()
        except httpx.HTTPStatusError as exc:
            if response.status_code == 404:
                raise APIError(
                    "Endpoint is not found. This may be due to an invalid model name. "
                    "See https://docs.periflow.ai/guides/serverless_endpoints/pricing "
                    "to find out availble models."
                ) from exc

            resp_content = response.read()
            raise APIError(resp_content.decode()) from exc


class AsyncServingAPI(BaseAPI[httpx.AsyncClient, _ProtoMsgType]):
    """Async Serving API interface."""

    def __init__(
        self,
        base_url: str,
        endpoint_id: Optional[str] = None,
        use_protobuf: bool = False,
        client: Optional[httpx.AsyncClient] = None,
    ) -> None:
        """Initializes AsyncServingAPI."""
        super().__init__(
            base_url=base_url, endpoint_id=endpoint_id, use_protobuf=use_protobuf
        )
        self._client = client or httpx.AsyncClient()

    async def _request(
        self, *, data: dict[str, Any], stream: bool, model: Optional[str] = None
    ) -> httpx.Response:
        # TODO: Add retry / handle timeout and etc.
        if self._endpoint_id is None and model is None:
            raise ValueError("`model` is required for serverless endpoints.")
        if self._endpoint_id is not None and model is not None:
            raise ValueError("`model` is not allowed for dedicated endpoints.")

        request = self._build_request(data=data, model=model)
        response = await self._client.send(request=request, stream=stream)
        await self._check_http_error(response)

        return response

    async def _check_http_error(self, response: httpx.Response) -> None:
        try:
            response.raise_for_status()
        except httpx.HTTPStatusError as exc:
            if response.status_code == 404:
                raise APIError(
                    "Endpoint is not found. This may be due to an invalid model name. "
                    "See https://docs.periflow.ai/guides/serverless_endpoints/pricing "
                    "to find out availble models."
                ) from exc

            resp_content = await response.aread()
            raise APIError(resp_content.decode()) from exc



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Friendli Checkpoint Quantizer Config Schema."""

from __future__ import annotations

from typing import Literal, Union

from pydantic import BaseModel, Field
from typing_extensions import Annotated

from friendli.enums import QuantDatasetFormat, QuantMode


class CalibrationDatasetConfig(BaseModel):
    """Calibration dataset config."""

    path_or_name: str = "lambada"
    format: QuantDatasetFormat = QuantDatasetFormat.JSON
    split: str = "validation"
    lookup_column_name: str = "text"
    num_samples: int = 128
    max_length: int = 512


class CommonQuantConfig(BaseModel):
    """Common quantization config."""

    mode: QuantMode
    device: str = "cuda:0"
    offload: bool = True
    seed: int = 42
    percentile: float = 99.9
    calibration_dataset: CalibrationDatasetConfig = Field(
        default_factory=CalibrationDatasetConfig
    )


class SmoothQuantArgs(BaseModel):
    """SmoothQuant args."""

    migration_strength: float = 0.5
    attn_fc_smoothing: bool = False
    ff2_smoothing: bool = False


class SmoothQuantConfig(CommonQuantConfig):
    """SmoothQuant config."""

    mode: Literal[QuantMode.SMOOTH_QUANT] = QuantMode.SMOOTH_QUANT
    smoothquant_args: SmoothQuantArgs = Field(default_factory=SmoothQuantArgs)


class AWQArgs(BaseModel):
    """AWQ args."""

    quant_bit: int = 4
    quant_group_size: int = 64


class AWQConfig(CommonQuantConfig):
    """AWQ config."""

    mode: Literal[QuantMode.AWQ] = QuantMode.AWQ
    awq_args: AWQArgs = Field(default_factory=AWQArgs)


OneOfQuantConfig = Annotated[
    Union[SmoothQuantConfig, AWQConfig], Field(discriminator="mode")
]


class QuantConfig(BaseModel):
    """Quantization config."""

    config: OneOfQuantConfig



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Friendli Logger."""

from __future__ import annotations

import logging
import os

_formatter = logging.Formatter()


class ColorFormatter(logging.Formatter):
    """Customized formatter with ANSI color."""

    grey = "\x1b[38;20m"
    yellow = "\x1b[33;20m"
    red = "\x1b[31;20m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"

    default_fmt = "%(asctime)s.%(msecs)05d: %(name)s %(levelname)s: %(message)s"
    default_datefmt = "%Y-%m-%d %H:%M:%S"

    FORMATS = {
        logging.DEBUG: grey + default_fmt + reset,
        logging.INFO: grey + default_fmt + reset,
        logging.WARNING: yellow + default_fmt + reset,
        logging.ERROR: red + default_fmt + reset,
        logging.CRITICAL: bold_red + default_fmt + reset,
    }

    def __init__(self):
        """Initialize CustomFormatter."""
        super().__init__(fmt=self.default_fmt, datefmt=self.default_datefmt)

        # Pre-create Formatter objects for each level to improve efficiency
        self.formatters = {
            level: logging.Formatter(fmt) for level, fmt in self.FORMATS.items()
        }

    def format(self, record):
        """Override format method."""
        formatter = self.formatters.get(record.levelno, self.formatters[logging.INFO])
        return formatter.format(record)


def get_logger(name: str) -> logging.Logger:
    """Get a formatted logger."""
    logger = logging.getLogger(name)

    handler = logging.StreamHandler()
    formatter = ColorFormatter()
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    logger.setLevel(os.environ.get("FRIENDLI_LOG_LEVEL", "INFO"))

    return logger


logger = get_logger("Friendli")
